"""
This module contains the classes relating to licenses.

See also: https://developer.github.com/v3/licenses/
"""
import base64

from . import models


class _License(models.GitHubCore):
    """Base license object."""

    class_name = "_License"

    def _update_attributes(self, license):
        self._api = license["url"]
        self.key = license["key"]
        self.name = license["name"]
        self.spdx_id = license["spdx_id"]

    def _repr(self):
        return f"<{self.class_name} [{self.name}]>"


class License(_License):
    """This object represents a license as returned by the GitHub API.

    See https://developer.github.com/v3/licenses/ for more information.

    This object has all of the attributes of :class:`ShortLicense` as well as
    the following attributes:

    .. attribute:: body

        The full text of this license.

    .. attribute:: conditions

        A list of the conditions of this license.

    .. attribute:: description

        The short description of this license.

    .. attribute:: featured

        A boolean attribute describing whether this license is featured on
        GitHub or not.

    .. attribute:: html_url

        The URL to view this license on GitHub.

    .. attribute:: implementation

        The short description of how a user applies this license to their
        original work.

    .. attribute:: limitations

        A list of limitations of this license.

    .. attribute:: permissions

        A list of the permissions granted by this license.
    """

    class_name = "License"

    def _update_attributes(self, license):
        super()._update_attributes(license)
        self.body = license["body"]
        self.conditions = license["conditions"]
        self.description = license["description"]
        self.featured = license["featured"]
        self.html_url = license["html_url"]
        self.implementation = license["implementation"]
        self.limitations = license["limitations"]
        self.permissions = license["permissions"]

    def _repr(self):
        return f"<License [{self.name}]>"


class ShortLicense(_License):
    """This object represents a license returned in a collection.

    GitHub's API returns different representations of objects in different
    contexts. This object reprsents a license that would be returned in a
    collection, e.g., retrieving all licenses from ``/licenses``.

    This object has the following attributes:

    .. attribute:: key

        The short, API name, for this license.

    .. attribute:: name

        The long form, legal name for this license.

    .. attribute:: spdx_id

        The Software Package Data Exchange (a.k.a, SPDX) identifier for this
        license.
    """

    class_name = "ShortLicense"
    _refresh_to = License


class RepositoryLicense(models.GitHubCore):
    """The representation of the repository's retrieved license.

    This object will be returned from
    :meth:`~github3.repos.repo.Repository.license` and behaves like
    :class:`~github3.repos.contents.Contents` with a few differences.
    This also includes a :attr:`license` attribute to access the licenses API.

    This object has the following attributes:

    .. attribute:: name

        The name of the file this license is stored in on the repository.

    .. attribute:: path

        The path to the file of this license in the repository.

    .. attribute:: sha

        The current SHA of this license file in the repository.

    .. attribute:: size

        The size in bytes of this file.

    .. attribute:: html_url

        The URL used to view this file in a browser.

    .. attribute:: git_url

        The URL to retrieve this license file via the git protocol.

    .. attribute:: download_url

        The URL used to download this license file from the repository.

    .. attribute:: type

        Analogous to :attr:`github3.repos.contents.Contents.type`, this should
        indicate whether this is a file, directory, or some other type.

    .. attribute:: content

        The content as returned by the API. This may be base64 encoded. See
        :meth:`decode_content` to retrieve the content in plain-text.

    .. attribute:: encoding

        The encoding of the content. For example, ``base64``.

    .. attribute:: links

        The dictionary of URLs returned in the ``_links`` key by the API.

    .. attribute:: license

        A :class:`github3.licenses.ShortLicense` instance representing the
        metadata GitHub knows about the license.
    """

    def _update_attributes(self, license):
        self._api = license["url"]
        self.name = license["name"]
        self.path = license["path"]
        self.sha = license["sha"]
        self.size = license["size"]
        self.html_url = license["html_url"]
        self.git_url = license["git_url"]
        self.download_url = license["download_url"]
        self.type = license["type"]
        self.content = license["content"]
        self.encoding = license["encoding"]
        self.links = license["_links"]
        self.license = ShortLicense(license["license"], self)

    def _repr(self):
        return f"<RepositoryLicense [{self.name}]>"

    def decode_content(self):
        """Decode the :attr:`content` attribute.

        If ``content`` is base64 encoded, decode this and return it.
        Otherwise, return ``content``.

        :returns:
            plain-text content of this license
        :rtype:
            text (str on Python 3)
        """
        if self.encoding == "base64":
            return base64.b64decode(self.content.encode("utf-8")).decode(
                "utf-8"
            )
        return self.content



"""This module contains the classes related to Events."""
import copy

from . import models


class EventUser(models.GitHubCore):
    """The class that represents the user information returned in Events.

    .. note::

        Refreshing this object will return a :class:`~github3.users.User`.

    .. attribute:: avatar_url

        The URL of the avatar image this user chose.

    .. attribute:: display_login

        The login that is displayed as part of the event.

    .. attribute:: gravatar_id

        The unique ID for the user's gravatar, if they're using gravatar to
        host their avatar.

    .. attribute:: id

        The user's unique ID in GitHub.

    .. attribute:: login

        The user's login (or handle) on GitHub.
    """

    def _update_attributes(self, user):
        self.avatar_url = user["avatar_url"]
        self.display_login = user.get("display_login")
        self.gravatar_id = user["id"]
        self.id = user["id"]
        self.login = user["login"]
        self._api = self.url = user["url"]

    def to_user(self):
        """Retrieve a full User object for this EventUser.

        :returns:
            The full information about this user.
        :rtype:
            :class:`~github3.users.User`
        """
        from . import users

        url = self._build_url("users", self.login)
        json = self._json(self._get(url), 200)
        return self._instance_or_null(users.User, json)

    refresh = to_user


class EventOrganization(models.GitHubCore):
    """Representation of the organization information returned in Events.

    .. note::

        Refreshing this object will return a
        :class:`~github3.orgs.Organization`.

    This object has the following attributes:

    .. attribute:: avatar_url

        The URL to this organization's avatar.

    .. attribute:: gravatar_id

        The unique identifier for this organization on Gravatar, if its
        avatar is hosted there.

    .. attribute:: id

        This organization's unique identifier on GitHub.

    .. attribute:: login

        The unique login for this organization.
    """

    def _update_attributes(self, org):
        self.avatar_url = org["avatar_url"]
        self.gravatar_id = org["id"]
        self.id = org["id"]
        self.login = org["login"]
        self._api = self.url = org["url"]

    def to_org(self):
        """Retrieve a full Organization object for this EventOrganization.

        :returns:
            The full information about this organization.
        :rtype:
            :class:`~github3.orgs.Organization`
        """
        from . import orgs

        url = self._build_url("orgs", self.login)
        json = self._json(self._get(url), 200)
        return self._instance_or_null(orgs.Organization, json)

    refresh = to_org


class EventPullRequest(models.GitHubCore):
    """Representation of a Pull Request returned in Events.

    .. note::

        Refreshing this object returns a :class:`~github3.pulls.PullRequest`.

    This object has the following attributes:

    .. attribute:: id

        The unique id of this pull request across all of GitHub.

    .. attribute:: number

        The number of this pull request on its repository.

    .. attribute:: state

        The state of this pull request during this event.

    .. attribute:: title

        The title of this pull request during this event.

    .. attribute:: locked

        A boolean attribute describing if this pull request was locked.
    """

    def _update_attributes(self, pull):
        self.id = pull["id"]
        self.number = pull["number"]
        self.state = pull["state"]
        self.title = pull["title"]
        self.locked = pull["locked"]
        self._api = self.url = pull["url"]

    def to_pull(self):
        """Retrieve a full PullRequest object for this EventPullRequest.

        :returns:
            The full information about this pull request.
        :rtype:
            :class:`~github3.pulls.PullRequest`
        """
        from . import pulls

        json = self._json(self._get(self.url), 200)
        return self._instance_or_null(pulls.PullRequest, json)

    refresh = to_pull


class EventReviewComment(models.GitHubCore):
    """Representation of review comments in events.

    .. note::

        Refreshing this object will return a new
        :class`~github3.pulls.ReviewComment`

    This object has the following attributes:

    .. attribute:: id

        The unique id of this comment across all of GitHub.

    .. attribute:: author_association

        The association the author has with this project.

    .. attribute:: body

        The markdown body of this review comment.

    .. attribute:: commit_id

        The identifier of the commit that this comment was left on.

    .. attribute:: created_at

        A :class:`~datetime.datetime` object representing the date and time
        this comment was created.

    .. attribute:: diff_hunk

        The section (or hunk) of the diff this comment was left on.

    .. attribute:: html_url

        The URL to view this comment in a browser.

    .. attribute:: links

        A dictionary of links to various items about this comment.

    .. attribute:: original_commit_id

        The identifier of original commit this comment was left on.

    .. attribute:: original_position

        The original position within the diff this comment was left.

    .. attribute:: path

        The path to the file this comment was left on.

    .. attribute:: position

        The current position within the diff this comment is placed.

    .. attribute:: pull_request_url

        The URL to retrieve the pull request informtation from the API.

    .. attribute:: updated_at

        A :class:`~datetime.datetime` object representing the date and time
        this comment was updated.

    .. attribute:: user

        A :class:`~github3.users.ShortUser` representing the user who authored
        this comment.
    """

    def _update_attributes(self, comment):
        from . import users

        self._api = comment["url"]
        self.id = comment["id"]
        self.author_association = comment["author_association"]
        self.body = comment["body"]
        self.commit_id = comment["commit_id"]
        self.created_at = self._strptime(comment["created_at"])
        self.diff_hunk = comment["diff_hunk"]
        self.html_url = comment["html_url"]
        self.links = comment["_links"]
        self.original_commit_id = comment["original_commit_id"]
        self.original_position = comment["original_position"]
        self.path = comment["path"]
        self.position = comment["position"]
        self.pull_request_url = comment["pull_request_url"]
        self.updated_at = self._strptime(comment["updated_at"])
        self.user = users.ShortUser(comment["user"], self)

    def to_review_comment(self):
        """Retrieve a full ReviewComment object for this EventReviewComment.

        :returns:
            The full information about this review comment
        :rtype:
            :class:`~github3.pulls.ReviewComment`
        """
        from . import pulls

        comment = self._json(self._get(self._api), 200)
        return pulls.ReviewComment(comment, self)

    refresh = to_review_comment


class EventIssue(models.GitHubCore):
    """The class that represents the issue information returned in Events."""

    def _update_attributes(self, issue):
        self.id = issue["id"]
        self.number = issue["number"]
        self.state = issue["state"]
        self.title = issue["title"]
        self.locked = issue["locked"]
        self._api = self.url = issue["url"]

    def to_issue(self):
        """Retrieve a full Issue object for this EventIssue."""
        from . import issues

        json = self._json(self._get(self.url), 200)
        return self._instance_or_null(issues.Issue, json)

    refresh = to_issue


class EventIssueComment(models.GitHubCore):
    """Representation of a comment left on an issue.

    See also: http://developer.github.com/v3/issues/comments/

    This object has the following attributes:

    .. attribute:: author_association

        The association of the author (:attr:`user`) with the repository
        this issue belongs to.

    .. attribute:: body

        The markdown formatted original text written by the author.

    .. attribute:: created_at

        A :class:`~datetime.datetime` object representing the date and time
        when this comment was created.

    .. attribute:: html_url

        The URL to view this comment in a browser.

    .. attribute:: id

        The unique identifier for this comment.

    .. attribute:: issue_url

        The URL of the parent issue in the API.

    .. attribute:: updated_at

        A :class:`~datetime.datetime` object representing the date and time
        when this comment was most recently updated.

    .. attribute:: user

        A :class:`~github3.users.ShortUser` representing the author of this
        comment.
    """

    def _update_attributes(self, comment):
        from . import users

        self._api = comment["url"]
        self.author_association = comment["author_association"]
        self.body = comment["body"]
        self.created_at = self._strptime(comment["created_at"])
        self.html_url = comment["html_url"]
        self.id = comment["id"]
        self.issue_url = comment["issue_url"]
        self.updated_at = self._strptime(comment["updated_at"])
        self.user = users.ShortUser(comment["user"], self)

    def to_issue_comment(self):
        """Retrieve the full IssueComment object for this comment.

        :returns:
            All the information about an IssueComment.
        :rtype:
            :class:`~github3.issues.comment.IssueComment`
        """
        from .issues import comment

        json = self._json(self._get(self.url), 200)
        return self._instance_or_null(comment.IssueComment, json)

    refresh = to_issue_comment


class Event(models.GitHubCore):
    """Represents an event as returned by the API.

    It structures and handles the data returned by via the
    `Events <https://developer.github.com/v3/activity/events>`_ section
    of the GitHub API.

    Two events can be compared like so::

        e1 == e2
        e1 != e2

    And that is equivalent to::

        e1.id == e2.id
        e1.id != e2.id

    .. attribute:: actor

        A :class:`~github3.events.EventUser` that represents the user whose
        action generated this event.

    .. attribute:: created_at

        A :class:`~datetime.datetime` representing when this event was created.

    .. attribute:: id

        The unique identifier for this event.

    .. attribute:: org

        If present, a :class:`~github3.events.EventOrganization` representing
        the organization on which this event occurred.

    .. attribute:: type

        The type of event this is.

        .. seealso::

            `Event Types Documentation`_
                GitHub's documentation of different event types

    .. attribute:: payload

        The payload of the event which has all of the details relevant to this
        event.

    .. attribute:: repo

        The string representation of the repository this event pertains to.

        .. versionchanged:: 1.0.0

            This restores the behaviour of the API. To get a tuple,
            representation, use ``self.repo.split('/', 1)``

    .. attribute:: public

        A boolean representing whether the event is publicly viewable or not.

    .. _Event Types Documentation:
        https://developer.github.com/v3/activity/events/types/
    """

    def _update_attributes(self, event):
        # If we don't copy this, then we end up altering _json_data which we do
        # not want to do:
        event = copy.deepcopy(event)
        self.actor = EventUser(event["actor"], self)
        self.created_at = self._strptime(event["created_at"])
        self.id = event["id"]
        self.org = event.get("org")
        if self.org:
            self.org = EventOrganization(event["org"], self)
        self.type = event["type"]
        handler = _payload_handlers.get(self.type, identity)
        self.payload = handler(event["payload"], self)
        self.repo = event["repo"]
        self.public = event["public"]

    def _repr(self):
        return f"<Event [{self.type[:-5]}]>"

    @staticmethod
    def list_types():
        """List available payload types."""
        return sorted(_payload_handlers.keys())


def _commitcomment(payload, session):
    from .repos.comment import ShortComment

    if payload.get("comment"):
        payload["comment"] = ShortComment(payload["comment"], session)
    return payload


def _follow(payload, session):
    if payload.get("target"):
        payload["target"] = EventUser(payload["target"], session)
    return payload


def _forkev(payload, session):
    from .repos import ShortRepository

    if payload.get("forkee"):
        payload["forkee"] = ShortRepository(payload["forkee"], session)
    return payload


def _gist(payload, session):
    from .gists import Gist

    if payload.get("gist"):
        payload["gist"] = Gist(payload["gist"], session)
    return payload


def _issuecomm(payload, session):
    if payload.get("issue"):
        payload["issue"] = EventIssue(payload["issue"], session)
    if payload.get("comment"):
        payload["comment"] = EventIssueComment(payload["comment"], session)
    return payload


def _issueevent(payload, session):
    if payload.get("issue"):
        payload["issue"] = EventIssue(payload["issue"], session)
    return payload


def _member(payload, session):
    if payload.get("member"):
        payload["member"] = EventUser(payload["member"], session)
    return payload


def _pullreqev(payload, session):
    if payload.get("pull_request"):
        payload["pull_request"] = EventPullRequest(
            payload["pull_request"], session
        )
    return payload


def _pullreqcomm(payload, session):
    # Transform the Pull Request attribute
    pull = payload.get("pull_request")
    if pull:
        payload["pull_request"] = EventPullRequest(pull, session)

    # Transform the Comment attribute
    comment = payload.get("comment")
    if comment:
        payload["comment"] = EventReviewComment(comment, session)
    return payload


def _release(payload, session):
    from .repos.release import Release

    release = payload.get("release")
    if release:
        payload["release"] = Release(release, session)
    return payload


def _team(payload, session):
    from .orgs import ShortTeam
    from .repos import ShortRepository

    if payload.get("team"):
        payload["team"] = ShortTeam(payload["team"], session)
    if payload.get("repo"):
        payload["repo"] = ShortRepository(payload["repo"], session)
    if payload.get("sender"):
        payload["sender"] = EventUser(payload["sender"], session)
    return payload


def identity(x, session):
    """Return the value."""
    return x


_payload_handlers = {
    "CommitCommentEvent": _commitcomment,
    "CreateEvent": identity,
    "DeleteEvent": identity,
    "FollowEvent": _follow,
    "ForkEvent": _forkev,
    "ForkApplyEvent": identity,
    "GistEvent": _gist,
    "GollumEvent": identity,
    "IssueCommentEvent": _issuecomm,
    "IssuesEvent": _issueevent,
    "MemberEvent": _member,
    "PublicEvent": identity,
    "PullRequestEvent": _pullreqev,
    "PullRequestReviewCommentEvent": _pullreqcomm,
    "PushEvent": identity,
    "ReleaseEvent": _release,
    "StatusEvent": identity,
    "TeamAddEvent": _team,
    "WatchEvent": identity,
}



"""Module containing the Issue logic."""
from json import dumps

from uritemplate import URITemplate

from . import comment
from . import event
from . import label
from . import milestone
from .. import models
from .. import users
from ..decorators import requires_auth


class _Issue(models.GitHubCore):
    """The :class:`Issue <Issue>` object.

    Please see GitHub's `Issue Documentation`_ for more information.

    .. _Issue Documentation:
        http://developer.github.com/v3/issues
    """

    class_name = "_Issue"

    def _update_attributes(self, issue):
        self._api = issue["url"]
        self.assignee = issue["assignee"]
        if self.assignee:
            self.assignee = users.ShortUser(self.assignee, self)
        self.assignees = issue["assignees"]
        if self.assignees:
            self.assignees = [
                users.ShortUser(assignee, self) for assignee in self.assignees
            ]
        self.body = issue["body"]
        self.closed_at = self._strptime(issue["closed_at"])
        self.comments_count = issue["comments"]
        self.comments_url = issue["comments_url"]
        self.created_at = self._strptime(issue["created_at"])
        self.events_url = issue["events_url"]
        self.html_url = issue["html_url"]
        self.id = issue["id"]
        self.labels_urlt = URITemplate(issue["labels_url"])
        self.locked = issue["locked"]
        self.milestone = issue["milestone"]
        if self.milestone:
            self.milestone = milestone.Milestone(self.milestone, self)
        self.number = issue["number"]
        self.original_labels = issue["labels"]
        if self.original_labels:
            self.original_labels = [
                label.ShortLabel(lbl, self) for lbl in self.original_labels
            ]
        self.pull_request_urls = issue.get("pull_request")
        self.state = issue["state"]
        self.title = issue["title"]
        self.updated_at = self._strptime(issue["updated_at"])
        self.user = users.ShortUser(issue["user"], self)

    def _repr(self):
        return "<{class_name} [#{n}]>".format(
            n=self.number, class_name=self.class_name
        )

    @requires_auth
    def add_assignees(self, users):
        """Assign ``users`` to this issue.

        This is a shortcut for :meth:`~github3.issues.issue.Issue.edit`.

        :param users:
            users or usernames to assign this issue to
        :type users:
            list of :class:`~github3.users.User`
        :type users:
            list of str
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        usernames = {getattr(user, "login", user) for user in users}
        assignees = list({a.login for a in self.assignees} | usernames)
        return self.edit(assignees=assignees)

    @requires_auth
    def add_labels(self, *args):
        """Add labels to this issue.

        :param str args:
            (required), names of the labels you wish to add
        :returns:
            list of labels
        :rtype:
            :class:`~github3.issues.label.ShortLabel`
        """
        url = self._build_url("labels", base_url=self._api)
        json = self._json(self._post(url, data=args), 200)
        return [label.ShortLabel(lbl, self) for lbl in json] if json else []

    @requires_auth
    def close(self):
        """Close this issue.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        assignee = self.assignee.login if self.assignee else ""
        number = self.milestone.number if self.milestone else None
        labels = [lbl.name for lbl in self.original_labels]

        return self.edit(
            self.title, self.body, assignee, "closed", number, labels
        )

    def comment(self, id_num):
        """Get a single comment by its id.

        The catch here is that id is NOT a simple number to obtain. If
        you were to look at the comments on issue #15 in
        sigmavirus24/Todo.txt-python, the first comment's id is 4150787.

        :param int id_num:
            (required), comment id, see example above
        :returns:
            the comment identified by ``id_num``
        :rtype:
            :class:`~github3.issues.comment.IssueComment`
        """
        json = None
        if int(id_num) > 0:  # Might as well check that it's positive
            base_url, _ = self.url.rsplit("/", 1)
            url = self._build_url("comments", str(id_num), base_url=base_url)
            json = self._json(self._get(url), 200)
        return self._instance_or_null(comment.IssueComment, json)

    def comments(self, number=-1, sort="", direction="", since=None):
        """Iterate over the comments on this issue.

        :param int number:
            (optional), number of comments to iterate over
            Default: -1 returns all comments
        :param str sort:
            accepted valuees: ('created', 'updated') api-default: created
        :param str direction:
            accepted values: ('asc', 'desc') Ignored without the sort parameter
        :param since:
            (optional), Only issues after this date will be returned. This can
            be a ``datetime`` or an ISO8601 formatted date string, e.g.,
            ``2012-05-20T23:10:27Z``
        :type since:
            datetime or string
        :returns:
            iterator of comments
        :rtype:
            :class:`~github3.issues.comment.IssueComment`
        """
        url = self._build_url("comments", base_url=self._api)
        params = comment.issue_comment_params(sort, direction, since)
        return self._iter(int(number), url, comment.IssueComment, params)

    @requires_auth
    def create_comment(self, body):
        """Create a comment on this issue.

        :param str body:
            (required), comment body
        :returns:
            the created comment
        :rtype:
            :class:`~github3.issues.comment.IssueComment`
        """
        json = None
        if body:
            url = self._build_url("comments", base_url=self._api)
            json = self._json(self._post(url, data={"body": body}), 201)
        return self._instance_or_null(comment.IssueComment, json)

    @requires_auth
    def edit(
        self,
        title=None,
        body=None,
        assignee=None,
        state=None,
        milestone=None,
        labels=None,
        assignees=None,
    ):
        """Edit this issue.

        :param str title:
            title of the issue
        :param str body:
            markdown formatted body (description) of the issue
        :param str assignee:
            login name of user the issue should be assigned to
        :param str state:
            accepted values: ('open', 'closed')
        :param int milestone:
            the number (not title) of the milestone to assign this to,
            or 0 to remove the milestone

            .. note::

                This is not the milestone's globally unique identifier, it's
                value in :attr:`~github3.issues.milestone.Milestone.number`.
        :param list labels:
            list of labels to apply this to
        :param assignees:
            (optional), login of the users to assign the issue to
        :type assignees:
            list of strings
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        json = None
        data = {
            "title": title,
            "body": body,
            "assignee": assignee,
            "state": state,
            "milestone": milestone,
            "labels": labels,
            "assignees": assignees,
        }
        self._remove_none(data)
        if data:
            if "milestone" in data and data["milestone"] == 0:
                data["milestone"] = None
            json = self._json(self._patch(self._api, data=dumps(data)), 200)
        if json:
            self._update_attributes(json)
            return True
        return False

    def events(self, number=-1):
        """Iterate over events associated with this issue only.

        :param int number:
            (optional), number of events to return. Default: -1 returns all
            events available.
        :returns:
            generator of events on this issues
        :rtype:
            :class:`~github3.issues.event.IssueEvent`
        """
        url = self._build_url("events", base_url=self._api)
        return self._iter(int(number), url, event.IssueEvent)

    def is_closed(self):
        """Check if the issue is closed.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        if self.closed_at or (self.state == "closed"):
            return True
        return False

    def labels(self, number=-1, etag=None):
        """Iterate over the labels associated with this issue.

        :param int number:
            (optional), number of labels to return. Default: -1 returns all
            labels applied to this issue.
        :param str etag:
            (optional), ETag from a previous request to the same endpoint
        :returns:
            generator of labels on this issue
        :rtype:
            :class:`~github3.issues.label.ShortLabel`
        """
        url = self._build_url("labels", base_url=self._api)
        return self._iter(int(number), url, label.ShortLabel, etag=etag)

    @requires_auth
    def lock(self):
        """Lock an issue.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("lock", base_url=self._api)
        return self._boolean(self._put(url), 204, 404)

    def pull_request(self):
        """Retrieve the pull request associated with this issue.

        :returns:
            the pull request associated with this issue
        :rtype:
            :class:`~github3.pulls.PullRequest`
        """
        from .. import pulls

        json = None
        pull_request_url = None
        if self.pull_request_urls is not None:
            pull_request_url = self.pull_request_urls.get("url")
        if pull_request_url:
            json = self._json(self._get(pull_request_url), 200)
        return self._instance_or_null(pulls.PullRequest, json)

    @requires_auth
    def remove_assignees(self, users):
        """Unassign ``users`` from this issue.

        This is a shortcut for :meth:`~github3.issues.issue.Issue.edit`.

        :param users:
            users or usernames to unassign this issue from
        :type users:
            list of :class:`~github3.users.User`
        :type users:
            list of str
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        usernames = {getattr(user, "login", user) for user in users}
        assignees = list({a.login for a in self.assignees} - usernames)
        return self.edit(assignees=assignees)

    @requires_auth
    def remove_label(self, name):
        """Remove label ``name`` from this issue.

        :param str name:
            (required), name of the label to remove
        :returns:
            list of removed labels
        :rtype:
            :class:`~github3.issues.label.ShortLabel`
        """
        url = self._build_url("labels", name, base_url=self._api)
        json = self._json(self._delete(url), 200, 404)
        labels = [label.ShortLabel(lbl, self) for lbl in json] if json else []
        return labels

    @requires_auth
    def remove_all_labels(self):
        """Remove all labels from this issue.

        :returns:
            the list of current labels (empty) if successful
        :rtype:
            list
        """
        # Can either send DELETE or [] to remove all labels
        return self.replace_labels([])

    @requires_auth
    def replace_labels(self, labels):
        """Replace all labels on this issue with ``labels``.

        :param list labels:
            label names
        :returns:
            list of labels
        :rtype:
            :class:`~github3.issues.label.ShortLabel`
        """
        url = self._build_url("labels", base_url=self._api)
        json = self._json(self._put(url, data=dumps(labels)), 200)
        return [label.ShortLabel(lbl, self) for lbl in json] if json else []

    @requires_auth
    def reopen(self):
        """Re-open a closed issue.

        .. note:: This is a short cut to using :meth:`edit`.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        assignee = self.assignee.login if self.assignee else ""
        number = self.milestone.number if self.milestone else None
        labels = [str(lbl) for lbl in self.original_labels]
        return self.edit(
            self.title, self.body, assignee, "open", number, labels
        )

    @requires_auth
    def unlock(self):
        """Unlock an issue.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("lock", base_url=self._api)
        return self._boolean(self._delete(url), 204, 404)


class Issue(_Issue):
    """Object for the full representation of an Issue.

    GitHub's API returns different amounts of information about issues based
    upon how that information is retrieved. This object exists to represent
    the full amount of information returned for a specific issue. For example,
    you would receive this class when calling
    :meth:`~github3.github.GitHub.issue`. To provide a clear
    distinction between the types of issues, github3.py uses different classes
    with different sets of attributes.

    .. versionchanged:: 1.0.0

    This object has all of the same attributes as a
    :class:`~github3.issues.issue.ShortIssue` as well as the following:

    .. attribute:: body_html

        The HTML formatted body of this issue.

    .. attribute:: body_text

        The plain-text formatted body of this issue.

    .. attribute:: closed_by

        If the issue is closed, a :class:`~github3.users.ShortUser`
        representing the user who closed the issue.
    """

    class_name = "Issue"

    def _update_attributes(self, issue):
        super()._update_attributes(issue)
        self.body_html = issue["body_html"]
        self.body_text = issue["body_text"]
        self.closed_by = issue["closed_by"]
        if self.closed_by:
            self.closed_by = users.ShortUser(self.closed_by, self)


class ShortIssue(_Issue):
    """Object for the shortened representation of an Issue.

    GitHub's API returns different amounts of information about issues based
    upon how that information is retrieved. Often times, when iterating over
    several issues, GitHub will return less information. To provide a clear
    distinction between the types of issues, github3.py uses different classes
    with different sets of attributes.

    .. versionadded:: 1.0.0

    This object has the following attributes:

    .. attribute:: assignee

        .. deprecated:: 1.0.0

            While the API still returns this attribute, it's not as useful in
            the context of multiple assignees.

        If a user is assigned to this issue, then it will be represented as a
        :class:`~github3.users.ShortUser`.

    .. attribute:: assignees

        If users are assigned to this issue, then they will be represented as
        a list of :class:`~github3.users.ShortUser`.

    .. attribute:: body

        The markdown formatted text of the issue as writen by the user who
        opened the issue.

    .. attribute:: closed_at

        If this issue is closed, this will be a :class:`~datetime.datetime`
        object representing the date and time this issue was closed. Otherwise
        it will be ``None``.

    .. attribute:: comments_count

        The number of comments on this issue.

    .. attribute:: comments_url

        The URL to retrieve the comments on this issue from the API.

    .. attribute:: created_at

        A :class:`~datetime.datetime` object representing the date and time
        this issue was created.

    .. attribute:: events_url

        The URL to retrieve the events related to this issue from the API.

    .. attribute:: html_url

        The URL to view this issue in a browser.

    .. attribute:: id

        The unique identifier for this issue in GitHub.

    .. attribute:: labels_urlt

        A :class:`~uritemplate.URITemplate` object that can expand to a URL to
        retrieve the labels on this issue from the API.

    .. attribute:: locked

        A boolean attribute representing whether or not this issue is locked.

    .. attribute:: milestone

        A :class:`~github3.issues.milestone.Milestone` object representing the
        milestone to which this issue was assigned.

    .. attribute:: number

        The number identifying this issue on its parent repository.

    .. attribute:: original_labels

        If any are assigned to this issue, the list of
        :class:`~github3.issues.label.ShortLabel` objects representing the
        labels returned by the API for this issue.

    .. attribute:: pull_request_urls

        If present, a dictionary of URLs for retrieving information about the
        associated pull request for this issue.

    .. attribute:: state

        The current state of this issue, e.g., ``'closed'`` or ``'open'``.

    .. attribute:: title

        The title for this issue.

    .. attribute:: updated_at

        A :class:`~datetime.datetime` object representing the date and time
        when this issue was last updated.

    .. attribute:: user

        A :class:`~github3.users.ShortUser` representing the user who opened
        this issue.
    """

    class_name = "ShortIssue"
    _refresh_to = Issue



"""This module contains the Contents object."""
from base64 import b64decode
from base64 import b64encode
from json import dumps

from .. import models
from ..decorators import requires_auth
from ..git import Commit


class Contents(models.GitHubCore):
    """A representation of file contents returned via the API.

    See also: http://developer.github.com/v3/repos/contents/

    This object has the following attributes:

    .. attribute:: content

        The body of the file. If this is present, it may be base64 encoded.

    .. attribute:: encoding

        The encoding used on the :attr:`content` when returning the data from
        the API, e.g., ``base64``. If :attr:`content` is not present this will
        not be present either.

    .. attribute:: decoded

        .. note:: This is a computed attribute which isn't returned by the API.
        .. versionchanged:: 0.5.2

        Decoded content of the file as a bytes object. If we try to decode
        to character set for you, we might encounter an exception which
        will prevent the object from being created. On python2 this is the
        same as a string, but on python3 you should call the decode method
        with the character set you wish to use, e.g.,
        ``content.decoded.decode('utf-8')``.

    .. attribute:: git_url

        The URL for the Git API pertaining to this file.

    .. attribute:: html_url

        The URL to open this file in a browser.

    .. attribute:: links

        A dictionary of links returned about the contents and related
        resources.

    .. attribute:: name

        The name of the file.

    .. attribute:: path

        The path to this file.

    .. attribute:: sha

        The SHA1 of the contents of this file.

    .. attribute:: size

        The size of file in bytes.

    .. attribute:: submodule_git_url

        The URL of the git submodule (if this is a git submodule).

    .. attribute:: target

        If the file is a symlink, this will be present and provides the type
        of file that the symlink points to.

    .. attribute:: type

        Type of content, e.g., ``'file'``, ``'symlink'``, or ``'submodule'``.
    """

    def _update_attributes(self, content):
        self._api = content["url"]
        self.content = content.get("content")
        self.encoding = content.get("encoding")
        self.decoded = self.content
        if self.encoding == "base64" and self.content is not None:
            self.decoded = b64decode(self.content.encode())
        self.download_url = content["download_url"]
        self.git_url = content["git_url"]
        self.html_url = content["html_url"]
        self.links = content["_links"]
        self.name = content["name"]
        self.path = content["path"]
        self._uniq = self.sha = content["sha"]
        self.size = content["size"]
        self.submodule_git_url = content.get("submodule_git_url")
        self.target = content.get("target")
        self.type = content["type"]

    def _repr(self):
        return f"<Contents [{self.path}]>"

    def __eq__(self, other):
        return self.decoded == other

    def __ne__(self, other):
        return self.sha != other

    @requires_auth
    def delete(self, message, branch=None, committer=None, author=None):
        """Delete this file.

        :param str message:
            (required), commit message to describe the removal
        :param str branch:
            (optional), branch where the file exists.
            Defaults to the default branch of the repository.
        :param dict committer:
            (optional), if no information is given the authenticated user's
            information will be used. You must specify both a name and email.
        :param dict author:
            (optional), if omitted this will be filled in with committer
            information. If passed, you must specify both a name and email.
        :returns:
            dictionary of new content and associated commit
        :rtype:
            :class:`~github3.repos.contents.Contents` and
            :class:`~github3.git.Commit`
        """
        json = {}
        if message:
            data = {
                "message": message,
                "sha": self.sha,
                "branch": branch,
                "committer": validate_commmitter(committer),
                "author": validate_commmitter(author),
            }
            self._remove_none(data)
            json = self._json(self._delete(self._api, data=dumps(data)), 200)
            if json and "commit" in json:
                json["commit"] = Commit(json["commit"], self)
            if json and "content" in json:
                json["content"] = self._instance_or_null(
                    Contents, json["content"]
                )
        return json

    @requires_auth
    def update(
        self, message, content, branch=None, committer=None, author=None
    ):
        """Update this file.

        :param str message:
            (required), commit message to describe the update
        :param str content:
            (required), content to update the file with
        :param str branch:
            (optional), branch where the file exists.
            Defaults to the default branch of the repository.
        :param dict committer:
            (optional), if no information is given the authenticated user's
            information will be used. You must specify both a name and email.
        :param dict author:
            (optional), if omitted this will be filled in with committer
            information. If passed, you must specify both a name and email.
        :returns:
            dictionary containing the updated contents object and the
            commit in which it was changed.
        :rtype:
            dictionary of :class:`~github3.repos.contents.Contents` and
            :class:`~github3.git.Commit`
        """
        if content and not isinstance(content, bytes):
            raise ValueError(  # (No coverage)
                "content must be a bytes object"
            )  # (No coverage)

        json = None
        if message and content:
            content = b64encode(content).decode("utf-8")
            data = {
                "message": message,
                "content": content,
                "branch": branch,
                "sha": self.sha,
                "committer": validate_commmitter(committer),
                "author": validate_commmitter(author),
            }
            self._remove_none(data)
            json = self._json(self._put(self._api, data=dumps(data)), 200)
            if json and "content" in json:
                self._update_attributes(json["content"])
                json["content"] = self
            if json and "commit" in json:
                json["commit"] = Commit(json["commit"], self)
        return json


def validate_commmitter(d):
    """Validate that there are enough details in the dictionary.

    When sending data to GitHub, we need to ensure we're sending the name and
    email for committer and author data.
    """
    if d and d.get("name") and d.get("email"):
        return d
    return None



"""This module contains all of the classes related to organizations."""
import typing as t
from json import dumps

from uritemplate import URITemplate

from . import models
from . import users
from .decorators import requires_auth
from .events import Event
from .projects import Project
from .repos import Repository
from .repos import ShortRepository

if t.TYPE_CHECKING:
    from . import users as _users


class ShortRepositoryWithPermissions(ShortRepository):
    class_name = "ShortRepositoryWithPermissions"

    def _update_attributes(self, repo) -> None:
        super()._update_attributes(repo)
        self.permissions = repo["permissions"]


class _Team(models.GitHubCore):
    """Base class for Team representations."""

    class_name = "_Team"
    # Roles available to members on a team.
    member_roles = frozenset(["member", "maintainer"])
    filterable_member_roles = member_roles.union(["all"])

    def _update_attributes(self, team):
        self._api = team["url"]
        self.id = team["id"]
        self.members_urlt = URITemplate(team["members_url"])
        self.name = team["name"]
        self.permission = team["permission"]
        self.privacy = team.get(
            "privacy"
        )  # TODO: Re-record cassettes to ensure this exists
        self.repositories_url = team["repositories_url"]
        self.slug = team["slug"]
        self.parent = None
        parent = team.get("parent")
        if parent:
            self.parent = ShortTeam(parent, self)

    def _repr(self):
        return "<{s.class_name} [{s.name}]>".format(s=self)

    @requires_auth
    def add_or_update_membership(self, username, role="member"):
        """Add or update the user's membership in this team.

        This returns a dictionary like so::

            {
                'state': 'pending',
                'url': 'https://api.github.com/teams/...',
                'role': 'member',
            }

        :param str username:
            (required), login of user whose membership is being modified
        :param str role:
            (optional), the role the user should have once their membership
            has been modified. Options: 'member', 'maintainer'. Default:
            'member'
        :returns:
            dictionary of the invitation response
        :rtype:
            dict
        """
        if role not in self.member_roles:
            raise ValueError(
                "'role' must be one of {}".format(
                    ", ".join(sorted(self.member_roles))
                )
            )
        data = {"role": role}
        url = self._build_url("memberships", username, base_url=self._api)
        return self._json(self._put(url, json=data), 200)

    @requires_auth
    def add_repository(self, repository, permission=""):
        """Add ``repository`` to this team.

        If a permission is not provided, the team's default permission
        will be assigned, by GitHub.

        :param str repository:
            (required), form: 'user/repo'
        :param str permission:
            (optional), ('pull', 'push', 'admin')
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        data = {}
        if permission:
            data = {"permission": permission}
        url = self._build_url("repos", repository, base_url=self._api)
        return self._boolean(self._put(url, data=dumps(data)), 204, 404)

    @requires_auth
    def delete(self):
        """Delete this team.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        return self._boolean(self._delete(self._api), 204, 404)

    @requires_auth
    def edit(
        self,
        name: str,
        permission: str = "",
        parent_team_id: t.Optional[int] = None,
        privacy: t.Optional[str] = None,
    ):
        """Edit this team.

        :param str name:
            (required), the new name of this team
        :param str permission:
            (optional), one of ('pull', 'push', 'admin')
            .. deprecated:: 3.0.0

                This was deprecated by the GitHub API.
        :param int parent_team_id:
            (optional), id of the parent team for this team
        :param str privacy:
            (optional), one of "closed" or "secret"
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        if name:
            data: t.Dict[str, t.Union[str, int]] = {"name": name}
            if permission:
                data["permission"] = permission
            if parent_team_id is not None:
                data["parent_team_id"] = parent_team_id
            if privacy in {"closed", "secret"}:
                data["privacy"] = privacy
            json = self._json(self._patch(self._api, data=dumps(data)), 200)
            if json:
                self._update_attributes(json)
                return True
        return False

    @requires_auth
    def has_repository(self, repository):
        """Check if this team has access to ``repository``.

        :param str repository:
            (required), form: 'user/repo'
        :returns:
            True if the team can access the repository, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("repos", repository, base_url=self._api)
        return self._boolean(self._get(url), 204, 404)

    @requires_auth
    def members(self, role=None, number=-1, etag=None):
        """Iterate over the members of this team.

        :param str role:
            (optional), filter members returned by their role in the team.
            Can be one of: ``"member"``, ``"maintainer"``, ``"all"``. Default:
            ``"all"``.
        :param int number:
            (optional), number of users to iterate over.  Default: -1 iterates
            over all values
        :param str etag:
            (optional), ETag from a previous request to the same endpoint
        :returns:
            generator of the members of this team
        :rtype:
            :class:`~github3.users.ShortUser`
        """
        headers = {}
        params = {}
        if role in self.filterable_member_roles:
            params["role"] = role
            headers["Accept"] = "application/vnd.github.ironman-preview+json"
        url = self._build_url("members", base_url=self._api)
        return self._iter(
            int(number),
            url,
            users.ShortUser,
            params=params,
            etag=etag,
            headers=headers,
        )

    @requires_auth
    def permissions_for(
        self, repository: str
    ) -> ShortRepositoryWithPermissions:
        headers = {"Accept": "application/vnd.github.v3.repository+json"}
        url = self._build_url("repos", repository, base_url=self._api)
        json = self._json(self._get(url, headers=headers), 200)
        return ShortRepositoryWithPermissions(json, self)

    @requires_auth
    def repositories(self, number=-1, etag=None):
        """Iterate over the repositories this team has access to.

        :param int number:
            (optional), number of repos to iterate over. Default: -1 iterates
            over all values
        :param str etag:
            (optional), ETag from a previous request to the same endpoint
        :returns:
            generator of repositories this team has access to
        :rtype:
            :class:`~github3.orgs.ShortRepositoryWithPermissions`
        """
        url = self._build_url("repos", base_url=self._api)
        return self._iter(
            int(number),
            url,
            ShortRepositoryWithPermissions,
            etag=etag,
        )

    @requires_auth
    def membership_for(self, username):
        """Retrieve the membership information for the user.

        :param str username:
            (required), name of the user
        :returns:
            dictionary with the membership
        :rtype:
            dict
        """
        url = self._build_url("memberships", username, base_url=self._api)
        json = self._json(self._get(url), 200)
        return json or {}

    @requires_auth
    def revoke_membership(self, username):
        """Revoke this user's team membership.

        :param str username:
            (required), name of the team member
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("memberships", username, base_url=self._api)
        return self._boolean(self._delete(url), 204, 404)

    @requires_auth
    def remove_repository(self, repository):
        """Remove ``repository`` from this team.

        :param str repository:
            (required), form: 'user/repo'
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("repos", repository, base_url=self._api)
        return self._boolean(self._delete(url), 204, 404)


class Team(_Team):
    """Object representing a team in the GitHub API.

    In addition to the attributes on a :class:`~github3.orgs.ShortTeam` a Team
    has the following attribute:

    .. attribute:: created_at

        A :class:`~datetime.datetime` instance representing the time and date
        when this team was created.

    .. attribute:: members_count

        The number of members in this team.

    .. attribute:: organization

        A :class:`~github3.orgs.ShortOrganization` representing the
        organization this team belongs to.

    .. attribute:: repos_count

        The number of repositories this team can access.

    .. attribute:: updated_at

        A :class:`~datetime.datetime` instance representing the time and date
        when this team was updated.

    Please see GitHub's `Team Documentation`_ for more information.

    .. _Team Documentation:
        http://developer.github.com/v3/orgs/teams/
    """

    class_name = "Team"

    def _update_attributes(self, team):
        super()._update_attributes(team)
        self.created_at = self._strptime(team["created_at"])
        self.members_count = team["members_count"]
        self.organization = ShortOrganization(team["organization"], self)
        self.repos_count = team["repos_count"]
        self.updated_at = self._strptime(team["updated_at"])


class ShortTeam(_Team):
    """Object representing a team in the GitHub API.

    .. attribute:: id

        Unique identifier for this team across all of GitHub.

    .. attribute:: members_count

        The number of members in this team.

    .. attribute:: members_urlt

        A :class:`~uritemplate.URITemplate` instance to either retrieve all
        members in this team or to test if a user is a member.

    .. attribute:: name

        The human-readable name provided to this team.

    .. attribute:: permission

        The level of permissions this team has, e.g., ``push``, ``pull``,
        or ``admin``.

    .. attribute:: privacy

        The privacy level of this team inside the organization.

    .. attribute:: repos_count

        The number of repositories this team can access.

    .. attribute:: repositories_url

        The URL of the resource to enumerate all repositories this team can
        access.

    .. attribute:: slug

        The handle for this team or the portion you would use in an
        at-mention after the ``/``, e.g., in ``@myorg/myteam`` the
        slug is ``myteam``.

    Please see GitHub's `Team Documentation`_ for more information.

    .. _Team Documentation:
        http://developer.github.com/v3/orgs/teams/
    """

    class_name = "ShortTeam"
    _refresh_to = Team


class _Organization(models.GitHubCore):
    """The :class:`Organization <Organization>` object.

    Please see GitHub's `Organization Documentation`_ for more information.

    .. _Organization Documentation:
        http://developer.github.com/v3/orgs/
    """

    class_name = "_Organization"

    # Filters available when listing members. Note: ``"2fa_disabled"``
    # is only available for organization owners.
    members_filters = frozenset(["2fa_disabled", "all"])

    # Roles available to members in an organization.
    member_roles = frozenset(["admin", "member"])
    filterable_member_roles = member_roles.union(["all"])

    # Roles for invitations, see also:
    # https://developer.github.com/v3/orgs/members/#create-organization-invitation
    invitation_roles = frozenset(
        ["admin", "direct_member", "billing_manager"]
    )

    def _update_attributes(self, org):
        self.avatar_url = org["avatar_url"]
        self.description = org["description"]
        self.events_url = org["events_url"]
        self.hooks_url = org["hooks_url"]
        self.id = org["id"]
        self.issues_url = org["issues_url"]
        self.login = org["login"]
        self.members_url = org["members_url"]
        self.public_members_urlt = URITemplate(org["public_members_url"])
        self.repos_url = org["repos_url"]
        self.url = self._api = org["url"]
        self.type = "Organization"

    def _repr(self):
        display_name = ""
        name = getattr(self, "name", None)
        if name is not None:
            display_name = f":{name}"
        return "<{s.class_name} [{s.login}{display}]>".format(
            s=self, display=display_name
        )

    @requires_auth
    def add_or_update_membership(self, username, role="member"):
        """Add a member or update their role.

        :param str username:
            (required), user to add or update.
        :param str role:
            (optional), role to give to the user. Options are ``member``,
            ``admin``. Defaults to ``member``.
        :returns:
            the created or updated membership
        :rtype:
            :class:`~github3.orgs.Membership`
        :raises:
            ValueError if role is not a valid choice
        """
        if role not in self.member_roles:
            raise ValueError(
                "'role' must be one of {}".format(
                    ", ".join(sorted(self.member_roles))
                )
            )
        data = {"role": role}
        url = self._build_url(
            "memberships", str(username), base_url=self._api
        )
        json = self._json(self._put(url, json=data), 200)
        return self._instance_or_null(Membership, json)

    @requires_auth
    def add_repository(self, repository, team_id):  # FIXME(jlk): add perms
        """Add ``repository`` to ``team``.

        .. versionchanged:: 1.0

            The second parameter used to be ``team`` but has been changed to
            ``team_id``. This parameter is now required to be an integer to
            improve performance of this method.

        :param str repository:
            (required), form: 'user/repo'
        :param int team_id:
            (required), team id
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        if int(team_id) < 0:
            return False

        url = self._build_url(
            "organizations",
            str(self.id),
            "team",
            str(team_id),
            "repos",
            str(repository),
        )
        return self._boolean(self._put(url), 204, 404)

    @requires_auth
    def blocked_users(
        self, number: int = -1, etag: t.Optional[str] = None
    ) -> t.Generator[users.ShortUser, None, None]:
        """Iterate over the users blocked by this organization.

        .. versionadded:: 2.1.0

        :param int number:
            (optional), number of users to iterate over.  Default: -1 iterates
            over all values
        :param str etag:
            (optional), ETag from a previous request to the same endpoint
        :returns:
            generator of the members of this team
        :rtype:
            :class:`~github3.users.ShortUser`
        """
        url = self._build_url("blocks", base_url=self._api)
        return self._iter(int(number), url, users.ShortUser, etag=etag)

    @requires_auth
    def block(self, username: users.UserLike) -> bool:
        """Block a specific user from an organization.

        .. versionadded:: 2.1.0

        :parameter str username:
            Name (or user-like instance) of the user to block.
        :returns:
            True if successful, Fales otherwise
        :rtype:
            bool
        """
        url = self._build_url("blocks", str(username), base_url=self._api)
        return self._boolean(self._put(url), 204, 404)

    @requires_auth
    def unblock(self, username: users.UserLike) -> bool:
        """Unblock a specific user from an organization.

        .. versionadded:: 2.1.0

        :parameter str username:
            Name (or user-like instance) of the user to unblock.
        :returns:
            True if successful, Fales otherwise
        :rtype:
            bool
        """
        url = self._build_url("blocks", str(username), base_url=self._api)
        return self._boolean(self._delete(url), 204, 404)

    @requires_auth
    def is_blocking(self, username: users.UserLike) -> bool:
        """Check if this organization is blocking a specific user.

        .. versionadded:: 2.1.0

        :parameter str username:
            Name (or user-like instance) of the user to unblock.
        :returns:
            True if successful, Fales otherwise
        :rtype:
            bool
        """
        url = self._build_url("blocks", str(username), base_url=self._api)
        return self._boolean(self._get(url), 204, 404)

    @requires_auth
    def create_hook(self, name, config, events=["push"], active=True):
        """Create a hook on this organization.

        :param str name:
            (required), name of the hook
        :param dict config:
            (required), key-value pairs which act as settings for this hook
        :param list events:
            (optional), events the hook is triggered for
        :param bool active:
            (optional), whether the hook is actually triggered
        :returns:
            the created hook
        :rtype:
            :class:`~github3.orgs.OrganizationHook`
        """
        json = None
        if name and config and isinstance(config, dict):
            url = self._build_url("hooks", base_url=self._api)
            data = {
                "name": name,
                "config": config,
                "events": events,
                "active": active,
            }
            json = self._json(self._post(url, data=data), 201)
        return OrganizationHook(json, self) if json else None

    @requires_auth
    def create_project(self, name, body=""):
        """Create a project for this organization.

        If the client is authenticated and a member of the organization, this
        will create a new project in the organization.

        :param str name:
            (required), name of the project
        :param str body:
            (optional), the body of the project
        :returns:
            the new project
        :rtype:
            :class:`~github3.projects.Project`
        """
        url = self._build_url("projects", base_url=self._api)
        data = {"name": name, "body": body}
        json = self._json(
            self._post(url, data, headers=Project.CUSTOM_HEADERS), 201
        )
        return self._instance_or_null(Project, json)

    @requires_auth
    def create_repository(
        self,
        name,
        description="",
        homepage="",
        private=False,
        has_issues=True,
        has_wiki=True,
        team_id=0,
        auto_init=False,
        gitignore_template="",
        license_template="",
        has_projects=True,
    ):
        """Create a repository for this organization.

        If the client is authenticated and a member of the organization, this
        will create a new repository in the organization.

            ``name`` should be no longer than 100 characters

        :param str name:
            (required), name of the repository

            .. warning:: this should be no longer than 100 characters
        :param str description:
            (optional)
        :param str homepage:
            (optional)
        :param bool private:
            (optional), If ``True``, create a private repository. API default:
            ``False``
        :param bool has_issues:
            (optional), If ``True``, enable issues for this repository. API
            default: ``True``
        :param bool has_wiki:
            (optional), If ``True``, enable the wiki for this repository. API
            default: ``True``
        :param int team_id:
            (optional), id of the team that will be granted access to this
            repository
        :param bool auto_init:
            (optional), auto initialize the repository.
        :param str gitignore_template:
            (optional), name of the template; this is ignored if auto_int is
            False.
        :param str license_template:
            (optional), name of the license; this is ignored if auto_int is
            False.
        :param bool has_projects:
            (optional), If ``True``, enable projects for this repository. API
            default: ``True``
        :returns:
            the created repository
        :rtype:
            :class:`~github3.repos.Repository`
        """
        url = self._build_url("repos", base_url=self._api)
        data = {
            "name": name,
            "description": description,
            "homepage": homepage,
            "private": private,
            "has_issues": has_issues,
            "has_wiki": has_wiki,
            "license_template": license_template,
            "auto_init": auto_init,
            "gitignore_template": gitignore_template,
            "has_projects": has_projects,
        }
        if int(team_id) > 0:
            data.update({"team_id": team_id})
        json = self._json(self._post(url, data), 201)
        return self._instance_or_null(Repository, json)

    @requires_auth
    def conceal_member(self, username):
        """Conceal ``username``'s membership in this organization.

        :param str username:
            username of the organization member to conceal
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("public_members", username, base_url=self._api)
        return self._boolean(self._delete(url), 204, 404)

    @requires_auth
    def create_team(
        self,
        name: str,
        repo_names: t.Optional[t.Sequence[str]] = [],
        maintainers: t.Optional[
            t.Union[t.Sequence[str], t.Sequence["_users._User"]]
        ] = [],
        permission: str = "pull",
        parent_team_id: t.Optional[int] = None,
        privacy: str = "secret",
    ):
        """Create a new team and return it.

        This only works if the authenticated user owns this organization.

        :param str name:
            (required), name to be given to the team
        :param list repo_names:
            (optional) repositories, e.g.  ['github/dotfiles']
        :param list maintainers:
            (optional) list of usernames who will be maintainers
        :param str permission:
            (optional), options:

            - ``pull`` -- (default) members can not push or administer
                repositories accessible by this team
            - ``push`` -- members can push and pull but not administer
                repositories accessible by this team
            - ``admin`` -- members can push, pull and administer
                repositories accessible by this team
        :param int parent_team_id:
            (optional), the ID of a team to set as the parent team.
        :param str privacy:
            (optional), options:

            - ``secret`` -- (default) only visible to organization
                owners and members of this team
            - ``closed`` -- visible to all members of this organization
        :returns:
            the created team
        :rtype:
            :class:`~github3.orgs.Team`
        """
        data: t.Dict[str, t.Union[t.List[str], str, int]] = {
            "name": name,
            "repo_names": [
                getattr(r, "full_name", r) for r in (repo_names or [])
            ],
            "maintainers": [
                getattr(m, "login", m) for m in (maintainers or [])
            ],
            "permission": permission,
            "privacy": privacy,
        }
        if parent_team_id:
            data.update({"parent_team_id": parent_team_id})

        url = self._build_url("teams", base_url=self._api)
        json = self._json(self._post(url, data), 201)
        return self._instance_or_null(Team, json)

    @requires_auth
    def edit(
        self,
        billing_email=None,
        company=None,
        email=None,
        location=None,
        name=None,
        description=None,
        has_organization_projects=None,
        has_repository_projects=None,
        default_repository_permission=None,
        members_can_create_repositories=None,
    ):
        """Edit this organization.

        :param str billing_email:
            (optional) Billing email address (private)
        :param str company:
            (optional)
        :param str email:
            (optional) Public email address
        :param str location:
            (optional)
        :param str name:
            (optional)
        :param str description:
            (optional) The description of the company.
        :param bool has_organization_projects:
            (optional) Toggles whether organization projects are enabled for
            the organization.
        :param bool has_repository_projects:
            (optional) Toggles whether repository projects are enabled for
            repositories that belong to the organization.
        :param string default_repository_permission:
            (optional) Default permission level members have for organization
            repositories:

            - ``read`` -- (default) can pull, but not push to or administer
                this repository.
            - ``write`` -- can pull and push, but not administer this
                repository.
            - ``admin`` -- can pull, push, and administer this repository.
            - ``none`` -- no permissions granted by default.
        :param bool members_can_create_repositories:
            (optional) Toggles ability of non-admin organization members to
            create repositories:

            - ``True`` -- (default) all organization members can create
                repositories.
            - ``False`` -- only admin members can create repositories.
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        json = None
        data = {
            "billing_email": billing_email,
            "company": company,
            "email": email,
            "location": location,
            "name": name,
            "description": description,
            "has_organization_projects": has_organization_projects,
            "has_repository_projects": has_repository_projects,
            "default_repository_permission": default_repository_permission,
            "members_can_create_repositories": members_can_create_repositories,
        }
        self._remove_none(data)

        if data:
            json = self._json(self._patch(self._api, data=dumps(data)), 200)

        if json:
            self._update_attributes(json)
            return True
        return False

    @requires_auth
    def hook(self, hook_id):
        """Get a single hook.

        :param int hook_id:
            (required), id of the hook
        :returns:
            the hook
        :rtype:
            :class:`~github3.orgs.OrganizationHook`
        """
        json = None
        if int(hook_id) > 0:
            url = self._build_url("hooks", str(hook_id), base_url=self._api)
            json = self._json(self._get(url), 200)
        return self._instance_or_null(OrganizationHook, json)

    @requires_auth
    def hooks(self, number=-1, etag=None):
        """Iterate over hooks registered on this organization.

        :param int number:
            (optional), number of hoks to return. Default: -1
            returns all hooks
        :param str etag:
            (optional), ETag from a previous request to the same endpoint
        :returns:
            generator of hooks
        :rtype:
            :class:`~github3.orgs.OrganizationHook`
        """
        url = self._build_url("hooks", base_url=self._api)
        return self._iter(int(number), url, OrganizationHook, etag=etag)

    @requires_auth
    def invite(
        self, team_ids, invitee_id=None, email=None, role="direct_member"
    ):
        """Invite the user to join this organization.

        :param list[int] team_ids:
            (required), list of team identifiers to invite the user to
        :param int invitee_id:
            (required if email is not specified), the identifier for the user
            being invited
        :param str email:
            (required if invitee_id is not specified), the email address of
            the user being invited
        :param str role:
            (optional) role to provide to the invited user. Must be one of
        :returns:
            the created invitation
        :rtype:
            :class:`~github3.orgs.Invitation`
        """
        if (invitee_id is None and email is None) or (
            invitee_id is not None and email is not None
        ):
            raise ValueError(
                "One of either 'invitee_id' or 'email' must be specified"
            )
        if not team_ids:
            raise ValueError(
                "'team_ids' must be a non-empty list of integers"
            )
        data = {"team_ids": team_ids}
        if invitee_id is not None:
            data["invitee_id"] = invitee_id
        else:
            data["email"] = email
        if role not in self.invitation_roles:
            raise ValueError(
                "'role' must be one of {}".format(
                    ", ".join(sorted(self.invitation_roles))
                )
            )
        headers = {"Accept": "application/vnd.github.dazzler-preview.json"}
        data["role"] = role
        url = self._build_url("invitations", base_url=self._api)
        json = self._json(self._post(url, data=data, headers=headers), 200)
        return self._instance_or_null(Invitation, json)

    @requires_auth
    def cancel_invite(self, invitee_id):
        """Cancel the invitation using ``invitee_id``
        of the user from the organization.

        :param int invitee_id:
            the identifier for the user being invited, to cancel its invitation
        :returns: bool
        """
        url = self._build_url("invitations", invitee_id, base_url=self._api)
        return self._boolean(self._delete(url), 204, 404)

    @requires_auth
    def failed_invitations(self):
        """List failed organization invitations.

        :returns: bool
        """
        url = self._build_url("failed_invitations", base_url=self._api)
        return self._json(self._get(url), 200, 404)

    def is_member(self, username):
        """Check if the user named ``username`` is a member.

        :param str username:
            name of the user you'd like to check
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("members", username, base_url=self._api)
        return self._boolean(self._get(url), 204, 404)

    def is_public_member(self, username):
        """Check if the user named ``username`` is a public member.

        :param str username:
            name of the user you'd like to check
        :returns:
            True if the user is a public member, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("public_members", username, base_url=self._api)
        return self._boolean(self._get(url), 204, 404)

    def all_events(self, username, number=-1, etag=None):
        """Iterate over all org events visible to the authenticated user.

        :param str username:
            (required), the username of the currently authenticated user.
        :param int number:
            (optional), number of events to return. Default: -1 iterates over
            all events available.
        :param str etag:
            (optional), ETag from a previous request to the same endpoint
        :returns:
            generator of events
        :rtype:
            :class:`~github3.events.Event`
        """
        url = self._build_url("users", username, "events", "orgs", self.login)
        return self._iter(int(number), url, Event, etag=etag)

    def public_events(self, number=-1, etag=None):
        """Iterate over public events for this org.

        :param int number:
            (optional), number of events to return. Default: -1 iterates over
            all events available.
        :param str etag:
            (optional), ETag from a previous request to the same
            endpoint
        :returns:
            generator of public events
        :rtype:
            :class:`~github3.events.Event`
        """
        url = self._build_url("events", base_url=self._api)
        return self._iter(int(number), url, Event, etag=etag)

    @requires_auth
    def invitations(self, number=-1, etag=None):
        """Iterate over outstanding invitations to this organization.

        :returns:
            generator of invitation objects
        :rtype:
            :class:`~github3.orgs.Invitation`
        """
        headers = {"Accept": "application/vnd.github.korra-preview"}
        url = self._build_url("invitations", base_url=self._api)
        return self._iter(
            int(number), url, Invitation, etag=etag, headers=headers
        )

    def members(self, filter=None, role=None, number=-1, etag=None):
        """Iterate over members of this organization.

        :param str filter:
            (optional), filter members returned by this method. Can be one of:
            ``"2fa_disabled"``, ``"all",``. Default: ``"all"``. Filtering by
            ``"2fa_disabled"`` is only available for organization owners with
            private repositories.
        :param str role:
            (optional), filter members returned by their role. Can be one of:
            ``"all"``, ``"admin"``, ``"member"``. Default: ``"all"``.
        :param int number:
            (optional), number of members to return. Default: -1 will return
            all available.
        :param str etag:
            (optional), ETag from a previous request to the same endpoint
        :returns:
            generator of members of this organization
        :rtype:
            :class:`~github3.users.ShortUser`
        """
        headers = {}
        params = {}
        if filter in self.members_filters:
            params["filter"] = filter
        if role in self.filterable_member_roles:
            params["role"] = role
            # TODO(sigmavirus24): Determine if the preview header is still
            # necessary
            headers["Accept"] = "application/vnd.github.ironman-preview+json"
        url = self._build_url("members", base_url=self._api)
        return self._iter(
            int(number),
            url,
            users.ShortUser,
            params=params,
            etag=etag,
            headers=headers,
        )

    @requires_auth
    def membership_for(self, username):
        """Obtain the membership status of ``username``.

        Implements
        https://developer.github.com/v3/orgs/members/#get-organization-membership

        :param str username:
            (required), username name of the user
        :returns:
            the membership information
        :rtype:
            :class:`~github3.orgs.Membership`
        """
        url = self._build_url("memberships", username, base_url=self._api)
        json = self._json(self._get(url), 200, 404)
        return self._instance_or_null(Membership, json)

    def public_members(self, number=-1, etag=None):
        """Iterate over public members of this organization.

        :param int number:
            (optional), number of members to return. Default: -1 will return
            all available.
        :param str etag:
            (optional), ETag from a previous request to the same endpoint
        :returns:
            generator of public members
        :rtype:
            :class:`~github3.users.ShortUser`
        """
        url = self._build_url("public_members", base_url=self._api)
        return self._iter(int(number), url, users.ShortUser, etag=etag)

    def project(self, id, etag=None):
        """Return the organization project with the given ID.

        :param int id:
            (required), ID number of the project
        :returns:
            requested project
        :rtype:
            :class:`~github3.projects.Project`
        """
        url = self._build_url("projects", id)
        json = self._json(self._get(url, headers=Project.CUSTOM_HEADERS), 200)
        return self._instance_or_null(Project, json)

    def projects(self, number=-1, etag=None):
        """Iterate over projects for this organization.

        :param int number:
            (optional), number of members to return. Default: -1 will return
            all available.
        :param str etag:
            (optional), ETag from a previous request to the same endpoint
        :returns:
            generator of organization projects
        :rtype:
            :class:`~github3.projects.Project`
        """
        url = self._build_url("projects", base_url=self._api)
        return self._iter(
            int(number),
            url,
            Project,
            etag=etag,
            headers=Project.CUSTOM_HEADERS,
        )

    @requires_auth
    def remove_membership(self, username):
        """Remove ``username`` from this organization.

        Unlike ``remove_member``, this will cancel a pending invitation.

        :param str username: (required), username of the member to remove
        :returns: bool
        """
        url = self._build_url("memberships", username, base_url=self._api)
        return self._boolean(self._delete(url), 204, 404)

    def repositories(self, type="", number=-1, etag=None):
        """Iterate over repos for this organization.

        :param str type:
            (optional), accepted values: ('all', 'public', 'member', 'private',
            'forks', 'sources'), API default: 'all'
        :param int number:
            (optional), number of members to return. Default: -1 will return
            all available.
        :param str etag:
            (optional), ETag from a previous request to the same endpoint
        :returns:
            generator of repositories in this organization
        :rtype:
            :class:`~github3.repos.Repository`
        """
        url = self._build_url("repos", base_url=self._api)
        params = {}
        if type in ("all", "public", "member", "private", "forks", "sources"):
            params["type"] = type
        return self._iter(int(number), url, ShortRepository, params, etag)

    @requires_auth
    def teams(self, number=-1, etag=None):
        """Iterate over teams that are part of this organization.

        :param int number:
            (optional), number of teams to return. Default: -1 returns all
            available teams.
        :param str etag:
            (optional), ETag from a previous request to the same endpoint
        :returns:
            generator of this organization's teams
        :rtype:
            :class:`~github3.orgs.ShortTeam`
        """
        url = self._build_url("teams", base_url=self._api)
        return self._iter(int(number), url, ShortTeam, etag=etag)

    @requires_auth
    def publicize_member(self, username: str) -> bool:
        """Make ``username``'s membership in this organization public.

        :param str username:
            the name of the user whose membership you wish to publicize
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("public_members", username, base_url=self._api)
        return self._boolean(self._put(url), 204, 404)

    @requires_auth
    def remove_member(self, username: str) -> bool:
        """Remove the user named ``username`` from this organization.

        .. note::

            Only a user may publicize their own membership. See also:
            https://developer.github.com/v3/orgs/members/#publicize-a-users-membership

        :param str username:
            name of the user to remove from the org
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("members", username, base_url=self._api)
        return self._boolean(self._delete(url), 204, 404)

    @requires_auth
    def remove_repository(
        self,
        repository: t.Union[Repository, ShortRepository, str],
        team_id: int,
    ):
        """Remove ``repository`` from the team with ``team_id``.

        :param str repository:
            (required), form: 'user/repo'
        :param int team_id:
            (required), the unique identifier of the team
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        if int(team_id) > 0:
            url = self._build_url(
                "organizations",
                str(self.id),
                "team",
                str(team_id),
                "repos",
                str(repository),
            )
            return self._boolean(self._delete(url), 204, 404)
        return False

    @requires_auth
    def team(self, team_id: int) -> t.Optional[Team]:
        """Return the team specified by ``team_id``.

        :param int team_id:
            (required), unique id for the team
        :returns:
            the team identified by the id in this organization
        :rtype:
            :class:`~github3.orgs.Team`
        """
        json = None
        if int(team_id) > 0:
            url = self._build_url(
                "organizations", str(self.id), "team", str(team_id)
            )
            json = self._json(self._get(url), 200)
        return self._instance_or_null(Team, json)

    @requires_auth
    def team_by_name(self, team_slug: str) -> t.Optional[Team]:
        """Return the team specified by ``team_slug``.

        :param str team_slug:
            (required), slug for the team
        :returns:
            the team identified by the slug in this organization
        :rtype:
            :class:`~github3.orgs.Team`
        """
        url = self._build_url("teams", str(team_slug), base_url=self._api)
        json = self._json(self._get(url), 200)
        return self._instance_or_null(Team, json)


class Organization(_Organization):
    """Object for the full representation of a Organization.

    GitHub's API returns different amounts of information about orgs based
    upon how that information is retrieved. This object exists to represent
    the full amount of information returned for a specific org. For example,
    you would receive this class when calling
    :meth:`~github3.github.GitHub.organization`. To provide a clear
    distinction between the types of orgs, github3.py uses different classes
    with different sets of attributes.

    .. versionchanged:: 1.0.0

    This object includes all attributes on
    :class:`~github3.orgs.ShortOrganization` as well as the following:

    .. attribute:: blog

        If set, the URL of this organization's blog.

    .. attribute:: company

        The name of the company that is associated with this organization.

    .. attribute:: created_at

        A :class:`~datetime.datetime` instance representing the time and date
        when this organization was created.

    .. attribute:: email

        The email address associated with this organization.

    .. attribute:: followers_count

        The number of users following this organization. Organizations no
        longer have followers so this number will always be 0.

    .. attribute:: following_count

        The number of users this organization follows. Organizations no
        longer follow users so this number will always be 0.

    .. attribute:: html_url

        The URL used to view this organization in a browser.

    .. attribute:: location

        The location of this organization, e.g., New York, NY.

    .. attribute:: name

        The display name of this organization.

    .. attribute:: public_repos_count

        The number of public repositories owned by thi sorganization.
    """

    class_name = "Organization"

    def _update_attributes(self, org):
        super()._update_attributes(org)
        self.created_at = self._strptime(org["created_at"])
        self.followers_count = org["followers"]
        self.following_count = org["following"]
        self.html_url = org["html_url"]
        self.public_repos_count = org["public_repos"]
        # GitHub just doesn't return these 4 attributes sometimes. Compare
        # /orgs/github3py to /orgs/testgh3py
        self.blog = org.get("blog")
        self.company = org.get("company")
        self.email = org.get("email")
        self.location = org.get("location")
        self.name = org.get("name")


class ShortOrganization(_Organization):
    """Object for the shortened representation of an Organization.

    GitHub's API returns different amounts of information about orgs based
    upon how that information is retrieved. Often times, when iterating over
    several orgs, GitHub will return less information. To provide a clear
    distinction between the types of orgs, github3.py uses different classes
    with different sets of attributes.

    .. versionadded:: 1.0.0

    .. attribute:: avatar_url

        The URL of the avatar image for this organization.

    .. attribute:: description

        The user-provided description of this organization.

    .. attribute:: events_url

        The URL to retrieve the events related to this organization.

    .. attribute:: hooks_url

        The URL for the resource to manage organization hooks.

    .. attribute:: id

        The unique identifier for this organization across GitHub.

    .. attribute:: issues_url

        The URL to retrieve the issues across this organization's
        repositories.

    .. attribute:: login

        The unique username of the organization.

    .. attribute:: members_url

        The URL to retrieve the members of this organization.

    .. attribute:: public_members_urlt

        A :class:`uritemplate.URITemplate` that can be expanded to either list
        the public members of this organization or verify a user is a public
        member.

    .. attribute:: repos_url

        The URL to retrieve the repositories in this organization.

    .. attribute:: url

        The URL to retrieve this organization from the GitHub API.

    .. attribute:: type

        .. deprecated:: 1.0.0

            This will be removed in a future release.

        Previously returned by the API to indicate the type of the account.
    """

    class_name = "ShortOrganization"
    _refresh_to = Organization


class Invitation(models.GitHubCore):
    """Object representing an invitation to an organization.

    .. attribute:: created_at

        A :class:`~datetime.datetime` instance representing the time and date
        when this invitation was created.

    .. attribute:: email

        The email address of the user invited to the organization.

    .. attribute:: id

        The unique identifier for this invitation.

    .. attribute:: invitation_team_url

        The API URL to retrieve the :class:`~github3.orgs.ShortTeam` objects
        associated with this invitation.

    .. attribute:: inviter

        A :class:`~github3.users.ShortUser` representing the user who invited
        the user identified by ``login``.

    .. attribute:: login

        The username of the user invited to the organization.

    .. attribute:: team_count

        The number of teams involved in this invitation.
    """

    def _update_attributes(self, json):
        self.created_at = self._strptime(json["created_at"])
        self.email = json["email"]
        self.id = json["id"]
        self.inviter = users.ShortUser(json["inviter"], self)
        self.login = json["login"]
        # NOTE(sigmavirus24): GitHub docs claim these should be present but
        # in testing it is not.
        self.invitation_team_url = json.get("invitation_team_url")
        self.team_count = json.get("team_count")

    def _repr(self):
        return "<Invitation {} for [{}] from [{}]>".format(
            self.id, self.login, self.inviter.login
        )

    @requires_auth
    def teams(self):
        """Retrieve the list of teams associated with this invite.

        :returns:
            generator of teams associated with this invitation
        :rtype:
            :class:`~github3.orgs.ShortTeam`
        """
        return self._iter(
            -1,
            self.invitation_team_url,
            ShortTeam,
            headers={"Accept": "application/vnd.github.dazzler-preview.json"},
        )


class Membership(models.GitHubCore):
    """Object describing a user's membership in teams and organizations.

    .. attribute:: organization

        A :class:`~github3.orgs.ShortOrganization` instance representing the
        organization this membership is part of.

    .. attribute:: organization_url

        The URL of the organization resource in the API that this membership
        is part of.

    .. attribute:: state

        The state of this membership, e.g., active or pending.

    .. attribute:: active

        .. warning::

            This is a computed attribute, it is not returned by the API.

        A boolean attribute equivalent to ``self.state.lower() == 'active'``.

    .. attribute:: pending

        .. warning::

            This is a computed attribute, it is not returned by the API.

        A boolean attribute equivalent to ``self.state.lower() == 'pending'``.
    """

    def _repr(self):
        return f"<Membership [{self.organization}]>"

    def _update_attributes(self, membership):
        self._api = membership["url"]
        self.organization = ShortOrganization(
            membership["organization"], self
        )
        self.organization_url = membership["organization_url"]
        self.role = membership["role"]
        self.state = membership["state"]
        self.user = users.ShortUser(membership["user"], self)

        self.active = self.state
        if self.active:
            self.active = self.state.lower() == "active"
        self.pending = self.state
        if self.pending:
            self.pending = self.state.lower() == "pending"

    @requires_auth
    def edit(self, state):
        """Edit the user's membership.

        :param str state:
            (required), the state the membership should be in. Only accepts
            ``"active"``.
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        json = None
        if state and state.lower() == "active":
            data = dumps({"state": state.lower()})
            json = self._json(self._patch(self._api, data=data), 200)
        if json is not None:
            self._update_attributes(json)
            return True
        return False


class OrganizationHook(models.GitHubCore):
    """The representation of a hook on an organization.

    See also: https://developer.github.com/v3/orgs/hooks/

    This object has the following attributes:

    .. attribute:: active

        A boolean attribute describing whether the hook is active or not.

    .. attribute:: config

        A dictionary containing the configuration for this hook.

    .. attribute:: created_at

        A :class:`~datetime.datetime` object representing the date and time
        when this hook was created.

    .. attribute:: events

        The list of events which trigger this hook.

    .. attribute:: id

        The unique identifier for this hook.

    .. attribute:: name

        The name provided to this hook.

    .. attribute:: updated_at

        A :class:`~datetime.datetime` object representing the date and time
        when this hook was updated.
    """

    def _update_attributes(self, hook, session=None):
        self._api = hook["url"]
        self.active = hook["active"]
        self.config = hook["config"]
        self.created_at = self._strptime(hook["created_at"])
        self.events = hook["events"]
        self.id = hook["id"]
        self.name = hook["name"]
        self.updated_at = self._strptime(hook["updated_at"])

    def _repr(self):
        return f"<OrganizationHook [{self.name}]>"

    @requires_auth
    def delete(self):
        """Delete this hook.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        return self._boolean(self._delete(self._api), 204, 404)

    @requires_auth
    def edit(self, config={}, events=[], active=True):
        """Edit this hook.

        :param dict config:
            (required), key-value pairs which act as settings for this hook
        :param list events:
            (optional), which events should this be triggered for
        :param bool active:
            (optional), should this event be active
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        data = {"config": config, "active": active}
        if events:
            data["events"] = events

        json = self._json(self._patch(self._api, data=dumps(data)), 200)

        if json:
            self._update_attributes(json)
            return True

        return False

    @requires_auth
    def ping(self):
        """Ping this hook.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("pings", base_url=self._api)
        return self._boolean(self._post(url), 204, 404)



"""Issue search results implementation."""
from ..issues import ShortIssue
from ..models import GitHubCore


class IssueSearchResult(GitHubCore):
    """A representation of a search result containing an issue.

    This object has the following attributes:

    .. attribute:: issue

        A :class:`~github3.issues.issue.ShortIssue` representing the issue
        found in this search result.

    .. attribute:: score

        The confidence score of this search result.

    .. attribute:: text_matches

        A list of matches in the issue for this search result.

        .. note::

            To receive these, you must pass ``text_match=True`` to
            :meth:`~github3.github.GitHub.search_issues`.
    """

    def _update_attributes(self, data):
        result = data.copy()
        self.score = result.pop("score")
        self.text_matches = result.pop("text_matches", [])
        self.issue = ShortIssue(result, self)

    def _repr(self):
        return f"<IssueSearchResult [{self.issue}]>"



"""Topics related logic."""
from .. import models


class Topics(models.GitHubCore):
    """Representation of the repository topics.

    .. attribute:: names

        The names of the topics.
    """

    def _update_attributes(self, topics):
        self.names = topics["names"]

    def _repr(self):
        return "<Topics [{}]>".format(", ".join(self.names))



"""
github3.api
===========

:copyright: (c) 2012-2014 by Ian Cordasco
:license: Modified BSD, see LICENSE for more details

"""
from .github import GitHub
from .github import GitHubEnterprise

gh = GitHub()


def login(username=None, password=None, token=None, two_factor_callback=None):
    """Construct and return an authenticated GitHub session.

    .. note::

        To allow you to specify either a username and password combination or
        a token, none of the parameters are required. If you provide none of
        them, you will receive ``None``.

    :param str username: login name
    :param str password: password for the login
    :param str token: OAuth token
    :param func two_factor_callback: (optional), function you implement to
        provide the Two-factor Authentication code to GitHub when necessary
    :returns: :class:`GitHub <github3.github.GitHub>`

    """
    g = None

    if (username and password) or token:
        g = GitHub()
        g.login(username, password, token, two_factor_callback)

    return g


def enterprise_login(
    username=None,
    password=None,
    token=None,
    url=None,
    two_factor_callback=None,
):
    """Construct and return an authenticated GitHubEnterprise session.

    .. note::

        To allow you to specify either a username and password combination or
        a token, none of the parameters are required. If you provide none of
        them, you will receive ``None``.

    :param str username: login name
    :param str password: password for the login
    :param str token: OAuth token
    :param str url: URL of a GitHub Enterprise instance
    :param func two_factor_callback: (optional), function you implement to
        provide the Two-factor Authentication code to GitHub when necessary
    :returns: :class:`GitHubEnterprise <github3.github.GitHubEnterprise>`

    """
    if not url:
        raise ValueError(
            "GitHub Enterprise requires you provide the URL of"
            " the instance"
        )

    g = None

    if (username and password) or token:
        g = GitHubEnterprise(url)
        g.login(username, password, token, two_factor_callback)

    return g



"""This module contains the RepoCommit classes."""
from . import status
from .. import checks
from .. import git
from .. import models
from .. import users
from .comment import RepoComment


class _RepoCommit(models.GitHubCore):
    """The :class:`RepoCommit <RepoCommit>` object.

    This represents a commit as
    viewed by a :class:`Repository`. This is different from a Commit object
    returned from the git data section.

    Two commit instances can be checked like so::

        c1 == c2
        c1 != c2

    And is equivalent to::

        c1.sha == c2.sha
        c1.sha != c2.sha

    """

    PREVIEW_HEADERS = {
        "Accept": "application/vnd.github.groot-preview+json,"
        "application/vnd.github.v3.full+json"
    }

    class_name = "_RepoCommit"

    def _update_attributes(self, commit):
        self._api = commit["url"]
        #: SHA of this commit.
        self._uniq = self.sha = commit["sha"]

    def _repr(self):
        return f"<{self.class_name} [{self.sha[:7]}]>"

    def check_runs(self):
        """Retrieve the check runs for this commit.

        .. versionadded:: 1.3.0

        :returns:
            the check runs for this commit
        :rtype:
            :class:`~github3.checks.CheckRun`
        """
        url = self._build_url("check-runs", base_url=self._api)
        return self._iter(
            -1,
            url,
            checks.CheckRun,
            headers=checks.CheckRun.CUSTOM_HEADERS,
            list_key="check_runs",
        )

    def check_suites(self):
        """Retrieve the check suites for this commit.

        .. versionadded:: 1.3.0

        :returns:
            the check suites for this commit
        :rtype:
            :class:`~github3.checks.CheckSuite`
        """
        url = self._build_url("check-suites", base_url=self._api)
        return self._iter(
            -1,
            url,
            checks.CheckSuite,
            headers=checks.CheckSuite.CUSTOM_HEADERS,
            list_key="check_suites",
        )

    def diff(self):
        """Retrieve the diff for this commit.

        :returns:
            the diff as a bytes object
        :rtype:
            bytes
        """
        resp = self._get(
            self._api, headers={"Accept": "application/vnd.github.diff"}
        )
        return resp.content if self._boolean(resp, 200, 404) else b""

    def patch(self):
        """Retrieve the patch formatted diff for this commit.

        :returns:
            the patch as a bytes object
        :rtype:
            bytes
        """
        resp = self._get(
            self._api, headers={"Accept": "application/vnd.github.patch"}
        )
        return resp.content if self._boolean(resp, 200, 404) else b""

    def status(self):
        """Retrieve the combined status for this commit.

        :returns:
            the combined status for this commit
        :rtype:
            :class:`~github3.repos.status.CombinedStatus`
        """
        url = self._build_url("status", base_url=self._api)
        json = self._json(self._get(url), 200)
        return self._instance_or_null(status.CombinedStatus, json)

    def statuses(self):
        """Retrieve the statuses for this commit.

        :returns:
            the statuses for this commit
        :rtype:
            :class:`~github3.repos.status.Status`
        """
        url = self._build_url("statuses", base_url=self._api)
        return self._iter(-1, url, status.Status)

    def comments(self, number=-1, etag=None):
        """Iterate over comments for this commit.

        :param int number:
            (optional), number of comments to return. Default: -1 returns all
            comments
        :param str etag:
            (optional), ETag from a previous request to the same endpoint
        :returns:
            generator of comments
        :rtype:
            :class:~github3.repos.comment.RepoComment`
        """
        url = self._build_url("comments", base_url=self._api)
        return self._iter(int(number), url, RepoComment, etag=etag)

    def associated_pull_requests(self, number=-1, etag=None):
        """Iterate pull requests associated with a commit.

        :param int number:
            (optional), number of comments to return. Default: -1 returns all
            pull requests
        :param str etag:
            (optional), ETag from a previous request to the same endpoint
        :returns:
            generator of pull requests
        :rtype:
            :class:~github3.pulls.PullRequest`
        """
        from .. import pulls

        url = self._build_url("pulls", base_url=self._api)
        return self._iter(
            number,
            url,
            pulls.ShortPullRequest,
            etag=etag,
            headers=self.PREVIEW_HEADERS,
        )


class RepoCommit(_RepoCommit):
    """Representation of a commit with repository and git data."""

    class_name = "Repository Commit"

    def _update_attributes(self, commit):
        super()._update_attributes(commit)
        #: The number of additions made in the commit.
        self.additions = 0
        #: The number of deletions made in the commit.
        self.deletions = 0
        #: The files that were modified by this commit.
        self.files = commit["files"]
        #: Total number of changes in the files.
        self.total = 0
        self.stats = commit["stats"]
        if self.stats:
            self.additions = self.stats["additions"]
            self.deletions = self.stats["deletions"]
            self.total = self.stats["total"]


class MiniCommit(_RepoCommit):
    """A commit returned on a ShortBranch."""

    class_name = "Mini Repository Commit"
    _refresh_to = RepoCommit


class ShortCommit(_RepoCommit):
    """Representation of an incomplete commit in a collection."""

    class_name = "Short Repository Commit"
    _refresh_to = RepoCommit

    def _update_attributes(self, commit):
        super()._update_attributes(commit)
        self.author = commit["author"]
        if self.author:
            self.author = users.ShortUser(self.author, self)
        self.comments_url = commit["comments_url"]
        self.commit = git.ShortCommit(commit["commit"], self)
        self.committer = commit["committer"]
        if self.committer:
            self.committer = users.ShortUser(self.committer, self)
        self.html_url = commit["html_url"]
        #: List of parents to this commit.
        self.parents = commit["parents"]
        #: The commit message
        self.message = getattr(self.commit, "message", None)



"""Module containing the GistFile object."""
from .. import models


class _GistFile(models.GitHubCore):
    """Base for GistFile classes."""

    def _update_attributes(self, gistfile):
        self.raw_url = gistfile["raw_url"]
        self.filename = gistfile["filename"]
        self.language = gistfile["language"]
        self.size = gistfile["size"]
        self.type = gistfile["type"]

    def _repr(self):
        return "<{s.class_name} [{s.filename}]>".format(s=self)

    def content(self):
        """Retrieve contents of file from key 'raw_url'.

        :returns:
            unaltered, untruncated contents of file.
        :rtype:
            bytes
        """
        resp = self._get(self.raw_url)
        if self._boolean(resp, 200, 404):
            return resp.content
        return None


class GistFile(_GistFile):
    """This represents the full file object returned by interacting with gists.

    The object has all of the attributes as returned by the API for a
    ShortGistFile as well as:

    .. attribute:: truncated

        A boolean attribute that indicates whether :attr:`original_content`
        contains all of the file's contents.

    .. attribute:: original_content

        The contents of the file (potentially truncated) returned by the API.
        If the file was truncated use :meth:`content` to retrieve it in its
        entirety.

    """

    class_name = "GistFile"

    def _update_attributes(self, gistfile):
        super()._update_attributes(gistfile)
        self.original_content = gistfile["content"]
        self.truncated = gistfile["truncated"]


class ShortGistFile(_GistFile):
    """This represents the file object returned by interacting with gists.

    The object has the following attributes as returned by the API for a Gist:

    .. attribute:: raw_url

        This URL provides access to the complete, untruncated content of the
        file represented by this object.

    .. attribute:: filename

        The string for the filename.

    .. attribute:: language

        The GitHub detected language for the file, e.g., Erlang, Python, text.

    .. attribute:: type

        The mime-type of the file. Related to :attr:`language`.

    .. attribute:: size

        The file size in bytes.
    """

    class_name = "ShortGistFile"
    _refresh_to = GistFile



"""This module contains the Gist, ShortGist, and GistFork objects."""
from json import dumps

from . import comment
from . import file as gistfile
from . import history
from .. import models
from .. import users
from ..decorators import requires_auth


class _Gist(models.GitHubCore):
    """This object holds all the information returned by Github about a gist.

    With it you can comment on or fork the gist (assuming you are
    authenticated), edit or delete the gist (assuming you own it).  You can
    also "star" or "unstar" the gist (again assuming you have authenticated).

    Two gist instances can be checked like so::

        g1 == g2
        g1 != g2

    And is equivalent to::

        g1.id == g2.id
        g1.id != g2.id

    See also: http://developer.github.com/v3/gists/

    """

    class_name = "_Gist"
    _file_class = gistfile.ShortGistFile

    def _update_attributes(self, gist):
        self.comments_count = gist["comments"]
        self.comments_url = gist["comments_url"]
        self.created_at = self._strptime(gist["created_at"])
        self.description = gist["description"]
        self.files = {
            filename: self._file_class(gfile, self)
            for filename, gfile in gist["files"].items()
        }
        self.git_pull_url = gist["git_pull_url"]
        self.git_push_url = gist["git_push_url"]
        self.html_url = gist["html_url"]
        self.id = gist["id"]
        self.owner = gist.get("owner")
        if self.owner is not None:
            self.owner = users.ShortUser(self.owner, self)
        self.public = gist["public"]
        self.updated_at = self._strptime(gist["updated_at"])
        self.url = self._api = gist["url"]

    def __str__(self):
        return self.id

    def _repr(self):
        return "<{s.class_name} [{s.id}]>".format(s=self)

    @requires_auth
    def create_comment(self, body):
        """Create a comment on this gist.

        :param str body:
            (required), body of the comment
        :returns:
            Created comment or None
        :rtype:
            :class:`~github3.gists.comment.GistComment`
        """
        json = None
        if body:
            url = self._build_url("comments", base_url=self._api)
            json = self._json(self._post(url, data={"body": body}), 201)
        return self._instance_or_null(comment.GistComment, json)

    @requires_auth
    def delete(self):
        """Delete this gist.

        :returns:
            Whether the deletion was successful or not
        :rtype:
            bool
        """
        return self._boolean(self._delete(self._api), 204, 404)

    @requires_auth
    def edit(self, description="", files={}):
        """Edit this gist.

        :param str description:
            (optional), description of the gist
        :param dict files:
            (optional), files that make up this gist; the key(s) should be the
            file name(s) and the values should be another (optional) dictionary
            with (optional) keys: 'content' and 'filename' where the former is
            the content of the file and the latter is the new name of the file.
        :returns:
            Whether the edit was successful or not
        :rtype:
            bool
        """
        data = {}
        json = None
        if description:
            data["description"] = description
        if files:
            data["files"] = files
        if data:
            json = self._json(self._patch(self._api, data=dumps(data)), 200)
        if json:
            self._update_attributes(json)
            return True
        return False

    @requires_auth
    def fork(self):
        """Fork this gist.

        :returns:
            New gist if successfully forked, ``None`` otherwise
        :rtype:
            :class:`~github3.gists.gist.ShortGist`
        """
        url = self._build_url("forks", base_url=self._api)
        json = self._json(self._post(url), 201)
        return self._instance_or_null(ShortGist, json)

    @requires_auth
    def is_starred(self):
        """Check to see if this gist is starred by the authenticated user.

        :returns:
            True if it is starred, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("star", base_url=self._api)
        return self._boolean(self._get(url), 204, 404)

    def comments(self, number=-1, etag=None):
        """Iterate over comments on this gist.

        :param int number:
            (optional), number of comments to iterate over.
            Default: -1 will iterate over all comments on the gist
        :param str etag:
            (optional), ETag from a previous request to the same endpoint
        :returns:
            generator of comments
        :rtype:
            :class:`~github3.gists.comment.GistComment`
        """
        url = self._build_url("comments", base_url=self._api)
        return self._iter(int(number), url, comment.GistComment, etag=etag)

    def commits(self, number=-1, etag=None):
        """Iterate over the commits on this gist.

        These commits will be requested from the API and should be the same as
        what is in ``Gist.history``.

        .. versionadded:: 0.6

        .. versionchanged:: 0.9

            Added param ``etag``.

        :param int number:
            (optional), number of commits to iterate over.
            Default: -1 will iterate over all commits associated with this
            gist.
        :param str etag:
            (optional), ETag from a previous request to this endpoint.
        :returns:
            generator of the gist's history
        :rtype:
            :class:`~github3.gists.history.GistHistory`
        """
        url = self._build_url("commits", base_url=self._api)
        return self._iter(int(number), url, history.GistHistory)

    def forks(self, number=-1, etag=None):
        """Iterator of forks of this gist.

        .. versionchanged:: 0.9

            Added params ``number`` and ``etag``.

        :param int number:
            (optional), number of forks to iterate over.
            Default: -1 will iterate over all forks of this gist.
        :param str etag:
            (optional), ETag from a previous request to this endpoint.
        :returns:
            generator of gists
        :rtype:
            :class:`~github3.gists.gist.ShortGist`
        """
        url = self._build_url("forks", base_url=self._api)
        return self._iter(int(number), url, ShortGist, etag=etag)

    @requires_auth
    def star(self):
        """Star this gist.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("star", base_url=self._api)
        return self._boolean(self._put(url), 204, 404)

    @requires_auth
    def unstar(self):
        """Un-star this gist.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("star", base_url=self._api)
        return self._boolean(self._delete(url), 204, 404)


class GistFork(models.GitHubCore):
    """This object represents a forked Gist.

    This has a subset of attributes of a
    :class:`~github3.gists.gist.ShortGist`:

    .. attribute:: created_at

        The date and time when the gist was created.

    .. attribute:: id

        The unique identifier of the gist.

    .. attribute:: owner

        The user who forked the gist.

    .. attribute:: updated_at

        The date and time of the most recent modification of the fork.

    .. attribute:: url

        The API URL for the fork.
    """

    def _update_attributes(self, fork):
        self.created_at = self._strptime(fork["created_at"])
        self.id = fork["id"]
        self.owner = users.ShortUser(fork["user"], self)
        self.updated_at = self._strptime(fork["updated_at"])
        self.url = self._api = fork["url"]

    def _repr(self):
        return f"<GistFork [{self.id}]>"

    def to_gist(self):
        """Retrieve the full Gist representation of this fork.

        :returns:
            The Gist if retrieving it was successful or ``None``
        :rtype:
            :class:`~github3.gists.gist.Gist`
        """
        json = self._json(self._get(self.url), 200)
        return self._instance_or_null(Gist, json)

    refresh = to_gist


class Gist(_Gist):
    """This object constitutes the full representation of a Gist.

    GitHub's API returns different amounts of information about gists
    based upon how that information is retrieved. This object exists to
    represent the full amount of information returned for a specific
    gist. For example, you would receive this class when calling
    :meth:`~github3.github.GitHub.gist`. To provide a clear distinction
    between the types of gists, github3.py uses different classes with
    different sets of attributes.

    This object has all the same attributes as
    :class:`~github3.gists.gist.ShortGist` as well as:

    .. attribute:: commits_url

        The URL to retrieve gist commits from the GitHub API.

    .. attribute:: original_forks

        A list of :class:`~github3.gists.gist.GistFork` objects representing
        each fork of this gist. To retrieve the most recent list of forks, use
        the :meth:`forks` method.

    .. attribute:: forks_url

        The URL to retrieve the current listing of forks of this gist.

    .. attribute:: history

        A list of :class:`~github3.gists.history.GistHistory` objects
        representing each change made to this gist.

    .. attribute:: truncated

        This is a boolean attribute that indicates whether the content of this
        Gist has been truncated or not.
    """

    class_name = "Gist"
    _file_class = gistfile.GistFile

    def _update_attributes(self, gist):
        super()._update_attributes(gist)
        self.commits_url = gist["commits_url"]
        self.original_forks = [GistFork(fork, self) for fork in gist["forks"]]
        self.forks_url = gist["forks_url"]
        self.history = [history.GistHistory(h, self) for h in gist["history"]]
        self.truncated = gist["truncated"]


class ShortGist(_Gist):
    """Short representation of a gist.

    GitHub's API returns different amounts of information about gists
    based upon how that information is retrieved. This object exists to
    represent the full amount of information returned for a specific
    gist. For example, you would receive this class when calling
    :meth:`~github3.github.GitHub.all_gists`. To provide a clear distinction
    between the types of gists, github3.py uses different classes with
    different sets of attributes.

    This object only has the following attributes:

    .. attribute:: url

        The GitHub API URL for this repository, e.g.,
        ``https://api.github.com/gists/6faaaeb956dec3f51a9bd630a3490291``.

    .. attribute:: comments_count

        Number of comments on this gist

    .. attribute:: description

        Description of the gist as written by the creator

    .. attribute:: html_url

        The URL of this gist on GitHub, e.g.,
        ``https://gist.github.com/sigmavirus24/6faaaeb956dec3f51a9bd630a3490291``

    .. attribute:: id

        The unique identifier for this gist.

    .. attribute:: public

        This is a boolean attribute  describing if the gist is public or
        private

    .. attribute:: git_pull_url

        The git URL to pull this gist, e.g.,
        ``git://gist.github.com/sigmavirus24/6faaaeb956dec3f51a9bd630a3490291.git``

    .. attribute:: git_push_url

        The git URL to push to gist, e.g.,
        ``git@gist.github.com/sigmavirus24/6faaaeb956dec3f51a9bd630a3490291.git``

    .. attribute:: created_at

        This is a datetime object representing when the gist was created.

    .. attribute:: updated_at
        This is a datetime object representing the last time this gist was
        most recently updated.

    .. attribute:: owner

        This attribute is a :class:`~github3.users.ShortUser` object
        representing the creator of the gist.

    .. attribute:: files

        A dictionary mapping the filename to a
        :class:`~github3.gists.gist.GistFile` object.

        .. versionchanged:: 1.0.0

            Previously this was a list but it has been converted to a
            dictionary to preserve the structure of the API.

    .. attribute:: comments_url

        The URL to retrieve the list of comments on the Gist via the API.
    """

    class_name = "ShortGist"
    _refresh_to = Gist



"""Module containing the logic for labels."""
from json import dumps

from ..decorators import requires_auth
from ..models import GitHubCore


class _Label(GitHubCore):
    SYMMETRA_PREVIEW_HEADERS = {
        "Accept": "application/vnd.github.symmetra-preview+json"
    }

    def _update_attributes(self, label):
        self._api = label["url"]
        self.color = label["color"]
        self.name = label["name"]
        self._uniq = self._api

    def _repr(self):
        return "<{0.class_name} [{0.name}]>".format(self)

    def __str__(self):
        return self.name

    @requires_auth
    def delete(self):
        """Delete this label.

        :returns:
            True if successfully deleted, False otherwise
        :rtype:
            bool
        """
        return self._boolean(self._delete(self._api), 204, 404)

    @requires_auth
    def update(self, name, color, description=None):
        """Update this label.

        :param str name:
            (required), new name of the label
        :param str color:
            (required), color code, e.g., 626262, no leading '#'
        :param str description:
            (optional), new description of the label
        :returns:
            True if successfully updated, False otherwise
        :rtype:
            bool
        """
        json = None

        if name and color:
            if color[0] == "#":
                color = color[1:]
            data = {"name": name, "color": color}
            if description is not None:
                data["description"] = description
            resp = self._patch(
                self._api,
                data=dumps(data),
                headers=self.SYMMETRA_PREVIEW_HEADERS,
            )
            json = self._json(resp, 200)

        if json:
            self._update_attributes(json)
            return True

        return False


class ShortLabel(_Label):
    """A representation of a label object defined on a repository.

    See also: http://developer.github.com/v3/issues/labels/

    This object has the following attributes::

    .. attribute:: color

        The hexadecimeal representation of the background color of this label.

    .. attribute:: name

        The name (display label) for this label.
    """

    class_name = "ShortLabel"


class Label(_Label):
    """A representation of a label object defined on a repository.

    See also: http://developer.github.com/v3/issues/labels/

    This object has the following attributes::

    .. attribute:: color

        The hexadecimeal representation of the background color of this label.

    .. attribute:: description

        The description for this label.

    .. attribute:: name

        The name (display label) for this label.
    """

    class_name = "Label"

    def _update_attributes(self, label):
        super()._update_attributes(label)
        self.description = label["description"]



"""A collection of useful utilities."""
import collections.abc as abc_collections
import datetime
import re

from requests import compat

# with thanks to https://code.google.com/p/jquery-localtime/issues/detail?id=4
ISO_8601 = re.compile(
    r"^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[0-1]|0"
    r"[1-9]|[1-2][0-9])(T(2[0-3]|[0-1][0-9]):([0-5][0-9]):([0"
    r"-5][0-9])(\.[0-9]+)?(Z|[+-](?:2[0-3]|[0-1][0-9]):[0-5]["
    r"0-9])?)?$"
)


def timestamp_parameter(timestamp, allow_none=True):
    """Function to check the conformance of timestamps passed by users.

    This will check that a string is a valid format and allow users to pass a
    datetime object which we will then convert to a proper ISO8601 date-time
    string.

    :param timestamp: string to be validated or datetime object to be
        converted.
    :param bool allow_none: whether or not to allow timestamp to be None.
        Default: ``True``
    :returns: valid ISO8601 string
    :rtype: str
    :raises: ValueError
    """
    if timestamp is None:
        if allow_none:
            return None
        raise ValueError("Timestamp value cannot be None")

    if isinstance(timestamp, datetime.datetime):
        return timestamp.isoformat() + "Z"

    if isinstance(timestamp, compat.basestring):
        if not ISO_8601.match(timestamp):
            raise ValueError(
                (
                    "Invalid timestamp: %s is not a valid ISO-8601"
                    " formatted date"
                )
                % timestamp
            )
        return timestamp

    raise ValueError("Cannot accept type %s for timestamp" % type(timestamp))


def stream_response_to_file(response, path=None):
    """Stream a response body to the specified file.

    Either use the ``path`` provided or use the name provided in the
    ``Content-Disposition`` header.

    :param response: A Response object from requests
    :type response: requests.models.Response
    :param str path: The full path and file name used to save the response
    :return: path to the file
    :rtype: str
    """
    pre_opened = False
    fd = None
    filename = None
    if path:
        if isinstance(getattr(path, "write", None), abc_collections.Callable):
            pre_opened = True
            fd = path
            filename = getattr(fd, "name", None)
        else:
            fd = open(path, "wb")
            filename = path
    else:
        header = response.headers["content-disposition"]
        i = header.find("filename=") + len("filename=")
        filename = header[i:]
        fd = open(filename, "wb")

    for chunk in response.iter_content(chunk_size=512):
        fd.write(chunk)

    if not pre_opened:
        fd.close()

    return filename



"""This module contains the Status object for GitHub's commit status API."""
from .. import models
from .. import users
from ..models import GitHubCore


class _Status(models.GitHubCore):
    """Representation of a status on a repository."""

    class_name = "_Status"

    def _update_attributes(self, status):
        self._api = status["url"]
        self.context = status["context"]
        self.created_at = self._strptime(status["created_at"])
        self.description = status["description"]
        self.id = status["id"]
        self.state = status["state"]
        self.target_url = status["target_url"]
        self.updated_at = self._strptime(status["updated_at"])

    def _repr(self):
        return "<{s.class_name} [{s.id}:{s.state}]>".format(s=self)


class Status(_Status):
    """Representation of a full status on a repository.

    See also: http://developer.github.com/v3/repos/statuses/

    This object has the same attributes as a
    :class:`~github3.repos.status.ShortStatus` as well as the following
    attributes:

    .. attribute:: creator

        A :class:`~github3.users.ShortUser` representing the user who created
        this status.
    """

    class_name = "Status"

    def _update_attributes(self, status):
        super()._update_attributes(status)
        self.creator = users.ShortUser(status["creator"], self)


class ShortStatus(_Status):
    """Representation of a short status on a repository.

    .. versionadded:: 1.0.0

    This is the representation found in a
    :class:`~github3.repos.status.CombinedStatus` object.

    See also: http://developer.github.com/v3/repos/statuses/

    This object has the following attributes:

    .. attribute:: context

        This is a string that explains the context of this status object.
        For example, ``'continuous-integration/travis-ci/pr'``.

    .. attribute:: created_at

        A :class:`~datetime.datetime` object representing the date and time
        when this status was created.

    .. attribute:: creator

        A :class:`~github3.users.ShortUser` representing the user who created
        this status.

    .. attribute:: description

        A short description of the status.

    .. attribute:: id

        The unique identifier of this status object.

    .. attribute:: state

        The state of this status, e.g., ``'success'``, ``'pending'``,
        ``'failure'``.

    .. attribute:: target_url

        The URL to retrieve more information about this status.

    .. attribute:: updated_at

        A :class:`~datetime.datetime` object representing the date and time
        when this status was most recently updated.
    """

    class_name = "ShortStatus"
    _refresh_to = Status


class CombinedStatus(GitHubCore):
    """A representation of the combined statuses in a repository.

    See also: http://developer.github.com/v3/repos/statuses/

    This object has the following attributes:

    .. attribute:: commit_url

        The URL of the commit this combined status is present on.

    .. attribute:: repository

        A :class:`~gitub3.repos.repo.ShortRepository` representing the
        repository on which this combined status exists.

    .. attribute:: sha

        The SHA1 of the commit this status exists on.

    .. attribute:: state

        The state of the combined status, e.g., ``'success'``, ``'pending'``,
        ``'failure'``.

    .. attribute:: statuses

        The list of :class:`~github3.repos.status.ShortStatus` objects
        representing the individual statuses that is combined in this object.

    .. attribute:: total_count

        The total number of sub-statuses.
    """

    def _update_attributes(self, combined_status):
        from . import repo

        self._api = combined_status["url"]
        self.commit_url = combined_status["commit_url"]
        self.repository = repo.ShortRepository(
            combined_status["repository"], self
        )
        self.sha = combined_status["sha"]
        self.state = combined_status["state"]
        statuses = combined_status["statuses"]
        self.statuses = [ShortStatus(s, self) for s in statuses]
        self.total_count = combined_status["total_count"]

    def _repr(self):
        f = "<CombinedStatus [{s.state}:{s.total_count} sub-statuses]>"
        return f.format(s=self)



"""This module contains the Authorization object."""
from .decorators import requires_basic_auth
from .models import GitHubCore


class Authorization(GitHubCore):
    """Representation of an OAuth Authorization.

    See also: https://developer.github.com/v3/oauth_authorizations/

    This object has the following attributes:

    .. attribute:: app

        Details about the application the authorization was created for.

    .. attribute:: created_at

        A :class:`~datetime.datetime` representing when this authorization was
        created.

    .. attribute:: fingerprint

        .. versionadded:: 1.0

        The optional parameter that is used to allow an OAuth application to
        create multiple authorizations for the same user. This will help
        distinguish two authorizations for the same app.

    .. attribute:: hashed_token

        .. versionadded:: 1.0

        This is the base64 of the SHA-256 digest of the token.

        .. seealso::

            `Removing Authorization Tokens`_
                The blog post announcing the removal of :attr:`token`.

    .. attribute:: id

        The unique identifier for this authorization.

    .. attribute:: note_url

        The URL that points to a longer description about the purpose of this
        autohrization.

    .. attribute:: note

        The short note provided when this authorization was created.

    .. attribute:: scopes

        The list of scopes assigned to this token.

        .. seealso::

            `Scopes for OAuth Applications`_
                GitHub's documentation around available scopes and what they
                mean

    .. attribute:: token

        If this authorization was created, this will contain the full token.
        Otherwise, this attribute will be an empty string.

    .. attribute:: token_last_eight

        .. versionadded:: 1.0

        The last eight characters of the token. This allows users to identify
        a token after the initial retrieval.

    .. attribute:: updated_at

        A :class:`~datetime.datetime` representing when this authorization was
        most recently updated.

    .. _Scopes for OAuth Applications:
        https://developer.github.com/apps/building-oauth-apps/scopes-for-oauth-apps/
    .. _Removing Authorization Tokens:
        https://developer.github.com/changes/2014-12-08-removing-authorizations-token/#what-should-you-do
    """

    def _update_attributes(self, auth):
        self._api = auth["url"]
        self.app = auth["app"]
        self.created_at = self._strptime(auth["created_at"])
        self.fingerprint = auth["fingerprint"]
        self.id = auth["id"]
        self.note_url = auth["note_url"]
        self.note = auth["note"]
        self.scopes = auth["scopes"]
        self.token = auth["token"]
        self.token_last_eight = auth["token_last_eight"]
        self.updated_at = self._strptime(auth["updated_at"])

    def _repr(self):
        return f"<Authorization [{self.name}]>"

    def _update(self, scopes_data, note, note_url):
        """Helper for add_scopes, replace_scopes, remove_scopes."""
        if note is not None:
            scopes_data["note"] = note
        if note_url is not None:
            scopes_data["note_url"] = note_url
        json = self._json(self._post(self._api, data=scopes_data), 200)

        if json:
            self._update_attributes(json)
            return True

        return False

    @requires_basic_auth
    def add_scopes(self, scopes, note=None, note_url=None):
        """Add the scopes to this authorization.

        .. versionadded:: 1.0

        :param list scopes:
            Adds these scopes to the ones present on this authorization
        :param str note:
            (optional), Note about the authorization
        :param str note_url:
            (optional), URL to link to when the user views the authorization
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        return self._update({"add_scopes": scopes}, note, note_url)

    @requires_basic_auth
    def delete(self):
        """Delete this authorization.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        return self._boolean(self._delete(self._api), 204, 404)

    @requires_basic_auth
    def remove_scopes(self, scopes, note=None, note_url=None):
        """Remove the scopes from this authorization.

        .. versionadded:: 1.0

        :param list scopes:
            Remove these scopes from the ones present on this authorization
        :param str note:
            (optional), Note about the authorization
        :param str note_url:
            (optional), URL to link to when the user views the authorization
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        return self._update({"rm_scopes": scopes}, note, note_url)

    @requires_basic_auth
    def replace_scopes(self, scopes, note=None, note_url=None):
        """Replace the scopes on this authorization.

        .. versionadded:: 1.0

        :param list scopes:
            Use these scopes instead of the previous list
        :param str note:
            (optional), Note about the authorization
        :param str note_url:
            (optional), URL to link to when the user views the authorization
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        return self._update({"scopes": scopes}, note, note_url)



"""Repository search results implementation."""
from .. import models
from .. import repos


class RepositorySearchResult(models.GitHubCore):
    """A representation of a search result containing a repository.

    This object has the following attributes::

    .. attribute:: repository

        A :class:`~github3.repos.repo.ShortRepository` representing the
        repository found by the search.

    .. attribute:: score

        The confidence score of this search result.

    .. attribute:: text_matches

        A list of the text matches in the repository that generated this
        result.

        .. note::

            To receive these, you must pass ``text_match=True`` to
            :meth:`~github3.github.GitHub.search_code`.
    """

    def _update_attributes(self, data):
        result = data.copy()
        self.score = result.pop("score")
        self.text_matches = result.pop("text_matches", [])
        self.repository = repos.ShortRepository(result, self)

    def _repr(self):
        return f"<RepositorySearchResult [{self.repository}]>"



"""The module containing deployment logic."""
from .. import users
from ..models import GitHubCore


class Deployment(GitHubCore):
    """Representation of a deployment of a repository at a point in time.

    See also: https://developer.github.com/v3/repos/deployments/

    This object has the following attributes:

    .. attribute:: created_at

        A :class:`~datetime.datetime` representing the date and time when this
        deployment was created.

    .. attribute:: creator

        A :class:`~github3.users.ShortUser` representing the user who created
        this deployment.

    .. attribute:: description

        The description of this deployment as provided by the :attr:`creator`.

    .. attribute:: environment

        The environment targeted for this deployment, e.g., ``'production'``,
        ``'staging'``.

    .. attribute:: id

        The unique identifier of this deployment.

    .. attribute:: payload

        The JSON payload string sent as part to trigger this deployment.

    .. attribute:: ref

        The reference used to create this deployment, e.g.,
        ``'deploy-20140526'``.

    .. attribute:: sha

        The SHA1 of the branch on GitHub when it was deployed.

    .. attribute:: statuses_url

        The URL to retrieve the statuses of this deployment from the API.

    .. attribute:: updated_at

        A :class:`~datetime.datetime` object representing the date and time
        when this deployment was most recently updated.
    """

    def _update_attributes(self, deployment):
        self._api = deployment["url"]
        self.created_at = self._strptime(deployment["created_at"])
        self.creator = users.ShortUser(deployment["creator"], self)
        self.description = deployment["description"]
        self.environment = deployment["environment"]
        self.id = deployment["id"]
        self.payload = deployment["payload"]
        self.ref = deployment["ref"]
        self.sha = deployment["sha"]
        self.statuses_url = deployment["statuses_url"]
        self.updated_at = self._strptime(deployment["updated_at"])

    def _repr(self):
        return f"<Deployment [{self.id} @ {self.sha}]>"

    def create_status(self, state, target_url=None, description=None):
        """Create a new deployment status for this deployment.

        :param str state:
            (required), The state of the status. Can be one of
            ``pending``, ``success``, ``error``, ``inactive``,
            ``in_progress``, ``queued``, or ``failure``.
        :param str target_url:
            The target URL to associate with this status.
            This URL should contain output to keep the user updated while the
            task is running or serve as historical information for what
            happened in the deployment. Default: ''.
        :param str description:
            A short description of the status. Default: ''.
        :return:
            the incomplete deployment status
        :rtype:
            :class:`~github3.repos.deployment.DeploymentStatus`
        """
        json = None

        if state in (
            "pending",
            "success",
            "error",
            "inactive",
            "in_progress",
            "queued",
            "failure",
        ):
            data = {
                "state": state,
                "target_url": target_url,
                "description": description,
            }
            self._remove_none(data)
            response = self._post(self.statuses_url, data=data)
            json = self._json(response, 201)

        return self._instance_or_null(DeploymentStatus, json)

    def statuses(self, number=-1, etag=None):
        """Iterate over the deployment statuses for this deployment.

        :param int number:
            (optional), the number of statuses to return.
            Default: -1, returns all statuses.
        :param str etag:
            (optional), the ETag header value from the last time
            you iterated over the statuses.
        :returns:
            generator of the statuses of this deployment
        :rtype:
            :class:`~github3.repos.deployment.DeploymentStatus`
        """
        i = self._iter(
            int(number), self.statuses_url, DeploymentStatus, etag=etag
        )
        return i


class DeploymentStatus(GitHubCore):
    """Representation of the status of a deployment of a repository.

    See also
    https://developer.github.com/v3/repos/deployments/#get-a-single-deployment-status

    This object has the following attributes:

    .. attribute:: created_at

        A :class:`~datetime.datetime` representing the date and time when this
        deployment status was created.

    .. attribute:: creator

        A :class:`~github3.users.ShortUser` representing the user who created
        this deployment status.

    .. attribute:: deployment_url

        The URL to retrieve the information about the deployment from the API.

    .. attribute:: description

        The description of this status as provided by the :attr:`creator`.

    .. attribute:: id

        The unique identifier of this deployment.

    .. attribute:: state

        The state of the deployment, e.g., ``'success'``.

    .. attribute:: target_url

        The URL to associate with this status. This should link to the output
        of the deployment.
    """

    def _update_attributes(self, status):
        self._api = status["url"]
        self.created_at = self._strptime(status["created_at"])
        self.creator = users.ShortUser(status["creator"], self)
        self.deployment_url = status["deployment_url"]
        self.description = status["description"]
        self.id = status["id"]
        self.state = status["state"]
        self.target_url = status["target_url"]
        self.updated_at = self._strptime(status["updated_at"])

    def _repr(self):
        return f"<DeploymentStatus [{self.id}]>"



"""GitHub Pages related logic."""
from .. import models


class PagesInfo(models.GitHubCore):
    """Representation of the information about a GitHub pages website.

    .. attribute:: cname

        The cname in use for the pages site, if one is set.

    .. attribute:: custom_404

        A boolean attribute indicating whether the user configured a custom
        404 page for this site.

    .. attribute:: status

        The current status of the pages site, e.g., ``built``.
    """

    def _update_attributes(self, info):
        self._api = info["url"]
        self.cname = info["cname"]
        self.custom_404 = info["custom_404"]
        self.status = info["status"]

    def _repr(self):
        info = self.cname or ""
        if info:
            info += "/"
        info += self.status or ""
        return f"<Pages Info [{info}]>"


class PagesBuild(models.GitHubCore):
    """Representation of a single build of a GitHub pages website.

    .. attribute:: commit

        The SHA of the commit that triggered this build.

    .. attribute:: created_at

        A :class:`~datetime.datetime` object representing the date and time
        when this build was created.

    .. attribute:: duration

        The time it spent processing this build.

    .. attribute:: error

        If this build errored, a dictionary containing the error message and
        details about the error.

    .. attribute:: pusher

        A :class:`~github3.users.ShortUser` representing the user who pushed
        the commit that triggered this build.

    .. attribute:: status

        The current statues of the build, e.g., ``building``.

    .. attribute:: updated_at

        A :class:`~datetime.datetime` object representing the date and time
        when this build was last updated.
    """

    def _update_attributes(self, build):
        from .. import users

        self._api = build["url"]
        self.commit = build["commit"]
        self.created_at = self._strptime(build["created_at"])
        self.duration = build["duration"]
        self.error = build["error"]
        self.pusher = users.ShortUser(build["pusher"], self)
        self.status = build["status"]
        self.updated_at = self._strptime(build["updated_at"])

    def _repr(self):
        return f"<Pages Build [{self.commit}/{self.status}]>"



from .. import users
from ..models import GitHubCore


class UserSearchResult(GitHubCore):
    """Representation of a search result for a user.

    This object has the following attributes:

    .. attribute:: score

        The confidence score of this result.

    .. attribute:: text_matches

        If present, a list of text strings that match the search string.

    .. attribute:: user

        A :class:`~github3.users.ShortUser` representing the user found
        in this search result.
    """

    def _update_attributes(self, data):
        result = data.copy()
        self.score = result.pop("score")
        self.text_matches = result.pop("text_matches", [])
        self.user = users.ShortUser(result, self)

    def _repr(self):
        return f"<UserSearchResult [{self.user}]>"



"""This module contains all the classes relating to projects."""
from json import dumps

from . import exceptions
from . import models
from . import pulls
from . import users
from .decorators import requires_auth
from .issues import issue


class Project(models.GitHubCore):
    """Object representing a single project from the API.

    See http://developer.github.com/v3/projects/ for more details.

    .. attribute:: body

        The Markdown formatted text describing the project.

    .. attribute:: created_at

        A :class:`~datetime.datetime` representing the date and time when
        this project was created.

    .. attribute:: creator

        A :class:`~github3.users.ShortUser` instance representing the user who
        created this project.

    .. attribute:: id

        The unique identifier for this project on GitHub.

    .. attribute:: name

        The name given to this project.

    .. attribute:: number

        The repository-local identifier of this project.

    .. attribute:: owner_url

        The URL of the resource in the API of the owning resource - either
        a repository or an organization.

    .. attribute:: updated_at

        A :class:`~datetime.datetime` representing the date and time when
        this project was last updated.
    """

    CUSTOM_HEADERS = {"Accept": "application/vnd.github.inertia-preview+json"}

    def _update_attributes(self, project):
        self._api = project["url"]
        self.body = project["body"]
        self.created_at = self._strptime(project["created_at"])
        self.creator = users.ShortUser(project["creator"], self)
        self.id = project["id"]
        self.name = project["name"]
        self.number = project["number"]
        self.owner_url = project["owner_url"]
        self.updated_at = self._strptime(project["updated_at"])

    def _repr(self):
        return f"<Project [#{self.id}]>"

    def column(self, id):
        """Get a project column with the given ID.

        :param int id:
            (required), the column ID
        :returns:
            the desired column in the project
        :rtype:
            :class:`~github3.projects.ProjectColumn`
        """
        url = self._build_url("projects", "columns", str(id))
        json = self._json(self._get(url, headers=Project.CUSTOM_HEADERS), 200)
        return self._instance_or_null(ProjectColumn, json)

    def columns(self, number=-1, etag=None):
        """Iterate over the columns in this project.

        :param int number:
            (optional), number of columns to return. Default: -1 returns all
            available columns.
        :param str etag:
            (optional), ETag from a previous request to the same endpoint
        :returns:
            generator of columns
        :rtype:
            :class:`~github3.project.ProjectColumn`
        """
        # TODO(sigmaviurs24): Determine if we need to construct from scratch
        # or if we can use `self._api` with 'columns' to build the URL
        url = self._build_url("projects", str(self.id), "columns")
        return self._iter(
            int(number),
            url,
            ProjectColumn,
            headers=Project.CUSTOM_HEADERS,
            etag=etag,
        )

    @requires_auth
    def create_column(self, name):
        """Create a column in this project.

        :param str name:
            (required), name of the column
        :returns:
            the created project column
        :rtype:
            :class:`~github3.projects.ProjectColumn`
        """
        url = self._build_url("columns", base_url=self._api)
        json = None
        if name:
            json = self._json(
                self._post(
                    url, data={"name": name}, headers=Project.CUSTOM_HEADERS
                ),
                201,
            )
        return self._instance_or_null(ProjectColumn, json)

    @requires_auth
    def delete(self):
        """Delete this project.

        :returns:
            True if successfully deleted, False otherwise
        :rtype:
            bool
        """
        return self._boolean(
            self._delete(self._api, headers=self.CUSTOM_HEADERS), 204, 404
        )

    @requires_auth
    def update(self, name=None, body=None):
        """Update this project.

        :param str name:
            (optional), new name of the project
        :param str body:
            (optional), new body of the project
        :returns:
            True if successfully updated, False otherwise
        :rtype:
            bool
        """
        data = {"name": name, "body": body}
        json = None
        self._remove_none(data)

        if data:
            json = self._json(
                self._patch(
                    self._api, data=dumps(data), headers=self.CUSTOM_HEADERS
                ),
                200,
            )

        if json:
            self._update_attributes(json)
            return True
        return False


class ProjectColumn(models.GitHubCore):
    """Object representing a column in a project.

    See http://developer.github.com/v3/projects/columns/

    .. attribute:: created_at

        A :class:`~datetime.datetime` object representing the date and time
        when the column was created.

    .. attribute:: id

        The unique identifier for this column across GitHub.

    .. attribute:: name

        The name given to this column.

    .. attribute:: project_url

        The URL used to retrieve the project that owns this column via the API.

    .. attribute:: updated_at

        A :class:`~datetime.datetime` object representing the date and time
        when the column was last updated.
    """

    def _update_attributes(self, project_column):
        self.created_at = self._strptime(project_column["created_at"])
        self.id = project_column["id"]
        self.name = project_column["name"]
        self.project_url = project_column["project_url"]
        self.updated_at = self._strptime(project_column["updated_at"])

    def _repr(self):
        return f"<ProjectColumn [#{self.id}]>"

    def card(self, id):
        """Get a project card with the given ID.

        :param int id:
            (required), the card ID
        :returns:
            the card identified by the provided id
        :rtype:
            :class:`~github3.projects.ProjectCard`
        """
        url = self._build_url("projects", "columns", "cards", str(id))
        json = self._json(self._get(url, headers=Project.CUSTOM_HEADERS), 200)
        return self._instance_or_null(ProjectCard, json)

    def cards(self, number=-1, etag=None):
        """Iterate over the cards in this column.

        :param int number:
            (optional), number of cards to return. Default: -1 returns all
            available cards.
        :param str etag:
            (optional), ETag from a previous request to the same endpoint
        :returns:
            generator of cards
        :rtype:
            :class:`~github3.project.ProjectCard`
        """
        url = self._build_url("projects", "columns", str(self.id), "cards")
        return self._iter(
            int(number),
            url,
            ProjectCard,
            headers=Project.CUSTOM_HEADERS,
            etag=etag,
        )

    @requires_auth
    def create_card_with_content_id(self, content_id, content_type):
        """Create a content card in this project column.

        :param int content_id:
            (required), the ID of the content
        :param str content_type:
            (required), the type of the content
        :returns:
            the created card
        :rtype:
            :class:`~github3.projects.ProjectCard`
        """
        if not content_id or not content_type:
            return None

        url = self._build_url("projects", "columns", str(self.id), "cards")
        json = None
        data = {"content_id": content_id, "content_type": content_type}
        json = self._json(
            self._post(url, data=data, headers=Project.CUSTOM_HEADERS), 201
        )
        return self._instance_or_null(ProjectCard, json)

    @requires_auth
    def create_card_with_issue(self, issue):
        """Create a card in this project column linked with an Issue.

        :param issue:
            (required), an issue with which to link the card. Can also be
            :class:`~github3.issues.ShortIssue`.
        :type issue:
            :class:`~github3.issues.Issue`
        :returns:
            the created card
        :rtype:
            :class:`~github3.projects.ProjectCard`
        """
        if not issue:
            return None
        return self.create_card_with_content_id(issue.id, "Issue")

    @requires_auth
    def create_card_with_note(self, note):
        """Create a note card in this project column.

        :param str note:
            (required), the note content
        :returns:
            the created card
        :rtype:
            :class:`~github3.projects.ProjectCard`
        """
        url = self._build_url("projects", "columns", str(self.id), "cards")
        json = None
        if note:
            json = self._json(
                self._post(
                    url, data={"note": note}, headers=Project.CUSTOM_HEADERS
                ),
                201,
            )
        return self._instance_or_null(ProjectCard, json)

    @requires_auth
    def delete(self):
        """Delete this column.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("projects", "columns", str(self.id))
        return self._boolean(
            self._delete(url, headers=Project.CUSTOM_HEADERS), 204, 404
        )

    @requires_auth
    def move(self, position):
        """Move this column.

        :param str position:
            (required), can be one of `first`, `last`, or `after:<column-id>`,
            where `<column-id>` is the id value of a column in the same
            project.
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        if not position:
            return False

        url = self._build_url("projects", "columns", str(self.id), "moves")
        data = {"position": position}
        return self._boolean(
            self._post(url, data=data, headers=Project.CUSTOM_HEADERS),
            201,
            404,
        )

    @requires_auth
    def update(self, name=None):
        """Update this column.

        :param str name:
            (optional), name of the column
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        data = {"name": name}
        json = None
        self._remove_none(data)

        if data:
            url = self._build_url("projects", "columns", str(self.id))
            json = self._json(
                self._patch(
                    url, data=dumps(data), headers=Project.CUSTOM_HEADERS
                ),
                200,
            )

        if json:
            self._update_attributes(json)
            return True
        return False


class ProjectCard(models.GitHubCore):
    """Object representing a "card" on a project.

    See http://developer.github.com/v3/projects/cards/

    .. attribute:: column_url

        The URL to retrieve this card's column via the API.

    .. attribute:: content_url

        The URl to retrieve this card's body content via the API.

    .. attribute:: created_at

        A :class:`~datetime.datetime` object representing the date and time
        when the column was created.

    .. attribute:: id

        The globally unique identifier for this card.

    .. attribute:: note

        The body of the note attached to this card.

    .. attribute:: updated_at

        A :class:`~datetime.datetime` object representing the date and time
        when the column was last updated.
    """

    def _update_attributes(self, project_card):
        #: The URL of this card's parent column
        self.column_url = project_card["column_url"]

        #: The URL of this card's associated content
        self.content_url = project_card.get("content_url")

        #: datetime object representing the last time the object was created
        self.created_at = project_card["created_at"]

        #: The ID of this card
        self.id = project_card["id"]

        #: The note attached to the card
        self.note = project_card["note"]

        #: datetime object representing the last time the object was changed
        self.updated_at = project_card["updated_at"]

    def _repr(self):
        return f"<ProjectCard [#{self.id}]>"

    @requires_auth
    def delete(self):
        """Delete this card.

        :returns:
            True if successfully deleted, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("projects", "columns", "cards", str(self.id))
        return self._boolean(
            self._delete(url, headers=Project.CUSTOM_HEADERS), 204, 404
        )

    @requires_auth
    def move(self, position, column_id):
        """Move this card.

        :param str position:
            (required), can be one of `top`, `bottom`, or `after:<card-id>`,
            where `<card-id>` is the id value of a card in the same column, or
            in the new column specified by `column_id`.
        :param int column_id:
            (required), the id value of a column in the same project.
        :returns:
            True if successfully moved, False
        :rtype:
            bool
        """
        if not position or not column_id:
            return False

        url = self._build_url(
            "projects", "columns", "cards", str(self.id), "moves"
        )
        data = {"position": position, "column_id": column_id}
        return self._boolean(
            self._post(url, data=data, headers=Project.CUSTOM_HEADERS),
            201,
            404,
        )

    @requires_auth
    def update(self, note=None):
        """Update this card.

        :param str note:
            (optional), the card's note content. Only valid for cards without
            another type of content, so this cannot be specified if the card
            already has a content_id and content_type.
        :returns:
            True if successfully updated, False otherwise
        :rtype:
            bool
        """
        data = {"note": note}
        json = None
        self._remove_none(data)

        if data:
            url = self._build_url(
                "projects", "columns", "cards", str(self.id)
            )
            json = self._json(
                self._patch(
                    url, data=dumps(data), headers=Project.CUSTOM_HEADERS
                ),
                200,
            )

        if json:
            self._update_attributes(json)
            return True
        return False

    def retrieve_issue_from_content(self):
        """Attempt to retrieve an Issue from the content url.

        :returns:
            The issue that backs up this particular project card if the card
            has a content_url.

            .. note::

                Cards can be created from Issues and Pull Requests. Pull
                Requests are also technically Issues so this method is always
                safe to call.
        :rtype:
            :class:`~github3.issues.issue.Issue`
        :raises:
            :class:`~github3.exceptions.CardHasNoContentUrl`
        """
        if self.content_url is None:
            raise exceptions.CardHasNoContentUrl(
                f"Card {self.id} has no content_url"
            )
        parsed = self._uri_parse(self.content_url)
        _, owner, repository, _, number = parsed.path[1:].split("/", 5)
        resp = self._get(
            self._build_url("repos", owner, repository, "issues", number)
        )
        json = self._json(resp, 200)
        return self._instance_or_null(issue.Issue, json)

    def retrieve_pull_request_from_content(self):
        """Attempt to retrieve an PullRequest from the content url.

        :returns:
            The pull request that backs this particular project card if the
            card has a content_url.

            .. note::

                Cards can be created from Issues and Pull Requests.
        :rtype:
            :class:`~github3.issues.issue.Issue`
        :raises:
            :class:`~github3.exceptions.CardHasNoContentUrl`
        """
        if self.content_url is None:
            raise exceptions.CardHasNoContentUrl(
                f"Card {self.id} has no content_url"
            )
        parsed = self._uri_parse(self.content_url)
        _, owner, repository, _, number = parsed.path[1:].split("/", 5)
        resp = self._get(
            self._build_url("repos", owner, repository, "pulls", number)
        )
        json = self._json(resp, 200)
        return self._instance_or_null(pulls.PullRequest, json)



"""Module containing session and auth logic."""
import collections.abc as abc_collections
import datetime
from contextlib import contextmanager
from logging import getLogger

import dateutil.parser
import requests

from . import __version__
from . import exceptions as exc

__url_cache__ = {}
__logs__ = getLogger(__package__)


def requires_2fa(response):
    """Determine whether a response requires us to prompt the user for 2FA."""
    if (
        response.status_code == 401
        and "X-GitHub-OTP" in response.headers
        and "required" in response.headers["X-GitHub-OTP"]
    ):
        return True
    return False


class BasicAuth(requests.auth.HTTPBasicAuth):
    """Sub-class requests's class so we have a nice repr."""

    def __repr__(self):
        """Use the username as the representation."""
        return f"basic {self.username}"


class TokenAuth(requests.auth.AuthBase):
    """Auth class that handles simple tokens."""

    header_format_str = "token {}"

    def __init__(self, token):
        """Store our token."""
        self.token = token

    def __repr__(self):
        """Return a nice view of the token in use."""
        return f"token {self.token[:4]}..."

    def __ne__(self, other):
        """Test for equality, or the lack thereof."""
        return not self == other

    def __eq__(self, other):
        """Test for equality, or the lack thereof."""
        return self.token == getattr(other, "token", None)

    def __call__(self, request):
        """Add the authorization header and format it."""
        request.headers["Authorization"] = self.header_format_str.format(
            self.token
        )
        return request


class GitHubSession(requests.Session):
    """Our slightly specialized Session object.

    Normally this is created automatically by
    :class:`~github3.github.GitHub`.  To use alternate values for
    network timeouts, this class can be instantiated directly and
    passed to the GitHub object.  For example:

    .. code-block:: python

       gh = github.GitHub(session=session.GitHubSession(
           default_connect_timeout=T, default_read_timeout=N))

    :param default_connect_timeout:
       the number of seconds to wait when establishing a connection to
       GitHub
    :type default_connect_timeout:
       float
    :param default_read_timeout:
       the number of seconds to wait for a response from GitHub
    :type default_read_timeout:
       float
    """

    auth = None
    __attrs__ = requests.Session.__attrs__ + [
        "base_url",
        "two_factor_auth_cb",
        "default_connect_timeout",
        "default_read_timeout",
        "request_counter",
    ]

    def __init__(self, default_connect_timeout=4, default_read_timeout=10):
        """Slightly modify how we initialize our session."""
        super().__init__()
        self.default_connect_timeout = default_connect_timeout
        self.default_read_timeout = default_read_timeout
        self.headers.update(
            {
                # Only accept JSON responses
                "Accept": "application/vnd.github.v3.full+json",
                # Only accept UTF-8 encoded data
                "Accept-Charset": "utf-8",
                # Always sending JSON
                "Content-Type": "application/json",
                # Set our own custom User-Agent string
                "User-Agent": f"github3.py/{__version__}",
            }
        )
        self.base_url = "https://api.github.com"
        self.two_factor_auth_cb = None
        self.request_counter = 0

    @property
    def timeout(self):
        """Return the timeout tuple as expected by Requests"""
        return (self.default_connect_timeout, self.default_read_timeout)

    def basic_auth(self, username, password):
        """Set the Basic Auth credentials on this Session.

        :param str username: Your GitHub username
        :param str password: Your GitHub password
        """
        if not (username and password):
            return

        self.auth = BasicAuth(username, password)

    def build_url(self, *args, **kwargs):
        """Build a new API url from scratch."""
        parts = [kwargs.get("base_url") or self.base_url]
        parts.extend(args)
        parts = [str(p) for p in parts]
        key = tuple(parts)
        __logs__.info("Building a url from %s", key)
        if key not in __url_cache__:
            __logs__.info("Missed the cache building the url")
            __url_cache__[key] = "/".join(parts)
        return __url_cache__[key]

    def handle_two_factor_auth(self, args, kwargs):
        """Handler for when the user has 2FA turned on."""
        headers = kwargs.pop("headers", {})
        headers.update({"X-GitHub-OTP": str(self.two_factor_auth_cb())})
        kwargs.update(headers=headers)
        return super().request(*args, **kwargs)

    def has_auth(self):
        """Check for whether or not the user has authentication configured."""
        return self.auth or self.headers.get("Authorization")

    def oauth2_auth(self, client_id, client_secret):
        """Use OAuth2 for authentication.

        It is suggested you install requests-oauthlib to use this.

        :param str client_id: Client ID retrieved from GitHub
        :param str client_secret: Client secret retrieved from GitHub
        """
        raise NotImplementedError("These features are not implemented yet")

    def request(self, *args, **kwargs):
        """Make a request, count it, and handle 2FA if necessary."""
        kwargs.setdefault("timeout", self.timeout)
        response = super().request(*args, **kwargs)
        self.request_counter += 1
        if requires_2fa(response) and self.two_factor_auth_cb:
            # No need to flatten and re-collect the args in
            # handle_two_factor_auth
            new_response = self.handle_two_factor_auth(args, kwargs)
            new_response.history.append(response)
            response = new_response
        return response

    def retrieve_client_credentials(self):
        """Return the client credentials.

        :returns: tuple(client_id, client_secret)
        """
        client_id = self.params.get("client_id")
        client_secret = self.params.get("client_secret")
        return (client_id, client_secret)

    def two_factor_auth_callback(self, callback):
        """Register our 2FA callback specified by the user."""
        if not callback:
            return

        if not isinstance(callback, abc_collections.Callable):
            raise ValueError("Your callback should be callable")

        self.two_factor_auth_cb = callback

    def token_auth(self, token):
        """Use an application token for authentication.

        :param str token: Application token retrieved from GitHub's
            /authorizations endpoint
        """
        if not token:
            return

        self.auth = TokenAuth(token)

    def app_bearer_token_auth(self, headers, expire_in):
        """Authenticate as an App to be able to view its metadata."""
        if not headers:
            return

        self.auth = AppBearerTokenAuth(headers, expire_in)

    def app_installation_token_auth(self, json):
        """Use an access token generated by an App's installation."""
        if not json:
            return

        self.auth = AppInstallationTokenAuth(
            json["token"], json["expires_at"]
        )

    @contextmanager
    def temporary_basic_auth(self, *auth):
        """Allow us to temporarily swap out basic auth credentials."""
        old_basic_auth = self.auth
        old_token_auth = self.headers.get("Authorization")

        self.basic_auth(*auth)
        yield

        self.auth = old_basic_auth
        if old_token_auth:
            self.headers["Authorization"] = old_token_auth

    @contextmanager
    def no_auth(self):
        """Unset authentication temporarily as a context manager."""
        old_basic_auth, self.auth = self.auth, None
        old_token_auth = self.headers.pop("Authorization", None)

        yield

        self.auth = old_basic_auth
        if old_token_auth:
            self.headers["Authorization"] = old_token_auth


def _utcnow():
    return datetime.datetime.now(dateutil.tz.UTC)


class AppInstallationTokenAuth(TokenAuth):
    """Use token authentication but throw an exception on expiration."""

    def __init__(self, token, expires_at):
        """Set-up our authentication handler."""
        super().__init__(token)
        self.expires_at_str = expires_at
        self.expires_at = dateutil.parser.parse(expires_at)

    def __repr__(self):
        """Return a nice view of the token in use."""
        return "app installation token {}... expiring at {}".format(
            self.token[:4], self.expires_at_str
        )

    @property
    def expired(self):
        """Indicate whether our token is expired or not."""
        now = _utcnow()
        return now > self.expires_at

    def __call__(self, request):
        """Add the authorization header and format it."""
        if self.expired:
            raise exc.AppInstallationTokenExpired(
                "Your app installation token expired at {}".format(
                    self.expires_at_str
                )
            )
        return super().__call__(request)


class AppBearerTokenAuth(TokenAuth):
    """Use JWT authentication but throw an exception on expiration."""

    header_format_str = "Bearer {}"

    def __init__(self, token, expire_in):
        """Set-up our authentication handler."""
        super().__init__(token)
        expire_in = datetime.timedelta(seconds=expire_in)
        self.expires_at = _utcnow() + expire_in

    def __repr__(self):
        """Return a helpful view of the token."""
        return "app bearer token {} expiring at {}".format(
            self.token[:4], str(self.expires_at)
        )

    @property
    def expired(self):
        """Indicate whether our token is expired or not."""
        now = _utcnow()
        return now > self.expires_at

    def __call__(self, request):
        """Add the authorization header and format it."""
        if self.expired:
            raise exc.AppTokenExpired(
                f"Your app token expired at {str(self.expires_at)}"
            )
        return super().__call__(request)



"""This module contains only the Hook object for GitHub's Hook API."""
from json import dumps

from ..decorators import requires_auth
from ..models import GitHubCore


class Hook(GitHubCore):
    """The representation of a hook on a repository.

    See also: http://developer.github.com/v3/repos/hooks/

    This object has the following attributes:

    .. attribute:: active

        A boolean attribute describing whether the hook is active or not.

    .. attribute:: config

        A dictionary containing the configuration for this hook.

    .. attribute:: created_at

        A :class:`~datetime.datetime` object representing the date and time
        when this hook was created.

    .. attribute:: events

        The list of events which trigger this hook.

    .. attribute:: id

        The unique identifier for this hook.

    .. attribute:: name

        The name provided to this hook.

    .. attribute:: updated_at

        A :class:`~datetime.datetime` object representing the date and time
        when this hook was updated.
    """

    def _update_attributes(self, hook, session=None):
        self._api = hook["url"]
        self.active = hook["active"]
        self.config = hook["config"]
        self.created_at = self._strptime(hook["created_at"])
        self.events = hook["events"]
        self.id = hook["id"]
        self.name = hook["name"]
        self.updated_at = self._strptime(hook["updated_at"])

    def _repr(self):
        return f"<Hook [{self.name}]>"

    @requires_auth
    def delete(self):
        """Delete this hook.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        return self._boolean(self._delete(self._api), 204, 404)

    @requires_auth
    def edit(
        self, config={}, events=[], add_events=[], rm_events=[], active=True
    ):
        """Edit this hook.

        :param dict config:
            (optional), key-value pairs of settings for this hook
        :param list events:
            (optional), which events should this be triggered for
        :param list add_events:
            (optional), events to be added to the list of events that this hook
            triggers for
        :param list rm_events:
            (optional), events to be removed from the list of events that this
            hook triggers for
        :param bool active:
            (optional), should this event be active
        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        data = {"config": config, "active": active}
        if events:
            data["events"] = events

        if add_events:
            data["add_events"] = add_events

        if rm_events:
            data["remove_events"] = rm_events

        json = self._json(self._patch(self._api, data=dumps(data)), 200)

        if json:
            self._update_attributes(json)
            return True

        return False

    @requires_auth
    def ping(self):
        """Ping this hook.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("pings", base_url=self._api)
        return self._boolean(self._post(url), 204, 404)

    @requires_auth
    def test(self):
        """Test this hook.

        :returns:
            True if successful, False otherwise
        :rtype:
            bool
        """
        url = self._build_url("tests", base_url=self._api)
        return self._boolean(self._post(url), 204, 404)
